MODULE Best_rational
IMPLICIT NONE
INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)


CONTAINS


SUBROUTINE cdfcor (nump, numq, numgr, ftble, pwr, iopt, epsq, indlw, altbl, &
                   indup, uptbl, indf, wtble, meq, mineq, alc, meqa, x,  &
                   error, qmin, iqmin, iter, indlp, iflag)
 
! Code converted using TO_F90 by Alan Miller
! Date: 2000-07-09  Time: 00:54:44
! Latest revision - 10 July 2000
 
!***BEGIN PROLOGUE  CDFCOR
!***DATE WRITTEN   720120 (YYMMDD)
!***REVISION DATE  880226   (YYMMDD)
!***CATEGORY NO.  E2C1
!***KEYWORDS  APPROXIMATION,CHEBYSHEV,RATIONAL,UNIFORM,BEST,
!             GENERALIZED,DIFFERENTIAL CORRECTION,MULTIVARIATE,
!             CONSTRAINTS,RESTRICTED RANGE,WEIGHTED
!***AUTHOR  KAUFMAN, EDWIN H. JR.,
!             DEPARTMENT OF MATHEMATICS
!             CENTRAL MICHIGAN UNIVERSITY
!             MOUNT PLEASANT, MICHIGAN 48859
!           TAYLOR, GERALD D.,
!             DEPARTMENT OF MATHEMATICS
!             COLORADO STATE UNIVERSITY
!             FORT COLLINS, COLORADO 80523
!***PURPOSE  THIS SUBROUTINE COMPUTES A BEST UNIFORM GENERALIZED
!            RATIONAL APPROXIMATION P/Q TO A FUNCTION F DEFINED
!            ON A FINITE SET OF GRID POINTS, WITH THE POSSIBILITY
!            OF INCLUDING A WEIGHT FUNCTION AND CONSTRAINTS ON
!            P/Q AND Q, AND ADDITIONAL ARBITRARY LINEAR CONSTRAINTS
!            ON THE COEFFICIENTS OF P AND Q.
!***DESCRIPTION
!
! THE VARIABLES IN THE CALLING SEQUENCE ARE AS FOLLOWS.  NONE OF
! THE STRICTLY INPUT VARIABLES IS CHANGED BY THIS PROGRAM.
!
!  NUMP   (INPUT) THIS IS THE NUMBER OF BASIS FUNCTIONS FOR THE NUMERATOR.
!
!  NUMQ   (INPUT) THIS IS THE NUMBER OF BASIS FUNCTIONS FOR THE DENOMINATOR.
!
!  NUMGR  (INPUT) THIS IS THE NUMBER OF GRID POINTS.
!
!  FTBLE  (INPUT) FTBLE(I) IS THE VALUE OF THE FUNCTION F AT GRID
!         POINT I (I=1,...,NUMGR).
!
!  PWR    (INPUT) PWR(I,J) IS THE VALUE OF THE JTH BASIS FUNCTION
!         (WHERE THE NUMERATOR BASIS FUNCTIONS PRECEDE THE
!         DENOMINATOR BASIS FUNCTIONS) AT GRID POINT I (J=1,...,
!         NUMP+NUMQ, I=1,...,NUMGR).
!
!  IOPT   (INPUT)  THIS IS THE OPTION SWITCH.  SETTING IOPT=0 TURNS
!         OFF ALL EXTRA OPTIONS.  THE PROPER SETTING OF IOPT TO
!         ACTIVATE ONE OR MORE OF THE EXTRA OPTIONS IS DISCUSSED IN
!         THE LONG DESCRIPTION BELOW.
!
!  EPSQ, INDLW, ALTBL, INDUP, UPTBL, INDF, WTBLE, MEQ, MINEQ, ALC,
!         MEQA  (OPTIONAL INPUT OR OUTPUT)
!         THESE ELEVEN VARIABLES ARE USED FOR INPUT OR OUTPUT FOR THE
!         EXTRA OPTIONS DESCRIBED IN THE LONG DESCRIPTION BELOW.  NONE
!         OF THEM NEED BE ASSIGNED VALUES IF IOPT=0.
!
!  X      (OUTPUT AND OPTIONAL INPUT) THE PROGRAM WILL PLACE THE
!         COMPUTED COEFFICIENT OF THE JTH BASIS FUNCTION IN X(J)
!         (J=1,...,NUMP+NUMQ).  X IS ALSO USED IF THE USER WISHES
!         TO SUPPLY AN INITIAL APPROXIMATION, IN WHICH CASE IOPT
!         MUST BE SET APPROPRIATELY (SEE LONG DESCRIPTION BELOW).
!
!  ERROR  (OUTPUT) THE PROGRAM WILL PLACE THE ERROR AT GRID POINT
!         I, WHICH WILL BE (F-P/Q)(GRID POINT I) IF NEITHER WEIGHT
!         FUNCTION OPTION IS USED (SEE LONG DESCRIPTION BELOW FOR
!         THE WEIGHT FUNCTION OPTIONS), IN ERROR(I) (I=1,...,NUMGR).
!         THE UNIFORM ERROR NORM WILL BE PLACED IN ERROR(NUMGR+1).
!
!  QMIN   (OUTPUT) THE PROGRAM WILL PLACE THE MINIMUM VALUE OF THE
!         DENOMINATOR ON THE GRID IN QMIN.
!
!  IQMIN  (OUTPUT) THE PROGRAM WILL PLACE THE INDEX OF THE GRID POINT
!         AT WHICH THE MINIMUM VALUE OF THE DENOMINATOR IS ACHIEVED IN IQMIN.
!
!  ITER   (OUTPUT) THE PROGRAM WILL PLACE THE NUMBER OF ITERATIONS
!         (NOT COUNTING INITIALIZATION) REQUIRED IN ITER.
!
!  INDLP  (OUTPUT) THE PROGRAM WILL PLACE INFORMATION ABOUT THE
!         PERFORMANCE OF THE LINEAR PROGRAMMING SUBROUTINE SLNPRO
!         IN THE TWO DIGIT VARIABLE INDLP (SEE COMMENTS AT THE END OF
!         THIS SUBROUTINE AND THE BEGINNING OF SUBROUTINE SLNPRO).
!         THE USER NORMALLY NEED NOT CONSIDER INDLP UNLESS IFLAG (SEE
!         BELOW) OR SOMETHING ELSE INDICATES SOMETHING IS WRONG.  IN
!         THAT CASE A CAREFUL STUDY OF THE INDICATED COMMENTS MAY
!         PROVIDE A CLUE AS TO WHAT HAPPENED.  FOR EXAMPLE, INDLP=44
!         USUALLY INDICATES INCORRECT INPUT TO THE PROGRAM.
!
!  IFLAG  (OUTPUT) IFLAG IS THE ERROR FLAG FOR THE PROGRAM.  IFLAG=0
!         MEANS THE PROGRAM APPEARED TO WORK NORMALLY, WITH NONE OF
!         THE CONDITIONS DESCRIBED BELOW FOR IFLAG .NE. O OCCURRING.
!         IFLAG POSITIVE IS A WARNING OF FAILURE IN THE INITIALIZATION.
!         IFLAG NEGATIVE INDICATES CAUTION SHOULD BE USED, ALTHOUGH
!         THE RETURNED APPROXIMATION MAY WELL BE ACCEPTABLE.  FOR
!         DETAILS CONCERNING SPECIFIC NONZERO VALUES OF IFLAG, SEE
!         THE LONG DESCRIPTION BELOW.
!
! THE DIMENSIONS OF THE ARRAYS IN THE CALLING SEQUENCE FOR CDFCOR ARE
!
! FTBLE(101), PWR(101,15), INDLW(101), ALTBL(101), INDUP(101),
! UPTBL(101), INDF(101), WTBLE(101), X(16), ERROR(102), ALC(101,16).
!
! IN ADDITION, IF SUBROUTINE STCOMP IS TO BE USED,
! DIMENSION T(202) MUST BE ADDED.
!
! THESE DIMENSIONS ARE SET SO THAT THE PROGRAM CAN BE APPLIED TO
! ANY PROBLEM WITH UP TO 15 BASIS FUNCTIONS (SO NUMP+NUMQ <= 15),
! UP TO 101 GRID POINTS (SO NUMGR <= 101), AND UP TO 101
! ADDITIONAL LINEAR CONSTRAINTS FROM OPTION 4 BELOW (SO
! MEQ+MINEQ <= 101), EVEN IF ALL THE EXTRA OPTIONS DESCRIBED
! IN THE LONG DESCRIPTION BELOW ARE ACTIVATED.  SEE THE LONG
! DESCRIPTION FOR THE DIMENSION PATTERN REQUIRED IN THE DRIVER
! PROGRAM AND SUBROUTINES IF MORE BASIS FUNCTIONS, GRID POINTS,
! OR ADDITIONAL LINEAR CONSTRAINTS ARE DESIRED, AND FOR THE
! DIMENSIONS OF THE INTERNAL WORK ARRAYS.
!
!***LONG DESCRIPTION
!
! THE USER CAN, BY SETTING IOPT AND THE CORRESPONDING OPTIONAL
! VARIABLES, USE ANY COMBINATION OF THE SEVEN EXTRA OPTIONS
! DESCRIBED BELOW.  AN OPTIONAL VARIABLE NEED BE ASSIGNED A VALUE
! BY THE USER ONLY IF ITS OPTION IS TO BE USED.  SETTING IOPT=0
! TURNS OFF ALL THE EXTRA OPTIONS.
!
!  OPTION 1  CONSTRAINTS (DENOMINATOR)
!
!            TO ACTIVATE THIS OPTION, SET THE ONES DIGIT OT IOPT
!            TO 1 (IF OPTION 4 IS NOT USED) OR TO 3 (IF OPTION 4 IS
!            USED) AND SET EPSQ AS FOLLOWS.
!
!  EPSQ   (INPUT) THIS IS THE CONSTANT LOWER BOUND ON THE VALUES
!         OF THE DENOMINATOR. THUS THE PROGRAM WILL SET UP
!         EXTRA CONSTRAINTS IN SUBROUTINE SLNPRO TO FORCE
!         Q >= EPSQ EVERYWHERE ON THE GRID.  THIS OPTION IS
!         USEFUL BECAUSE THE COMPUTED APPROXIMATION P/Q MAY BE OF
!         MORE VALUE IF ITS DENOMINATOR IS KEPT FARTHER AWAY FROM
!         ZERO, AND ALSO BECAUSE IN DIFFICULT CASES THIS OPTION
!         MAY ALLEVIATE NUMERICAL DIFFICULTIES IN THE PROGRAM.
!
!  OPTION 2  CONSTRAINTS (LOWER RESTRICTED RANGE)
!
!            TO ACTIVATE THIS OPTION, SET THE TENS DIGIT OF IOPT
!            TO 1 AND SET INDLW AND ALTBL AS FOLLOWS.
!
!  INDLW  (INPUT) INDLW(I)=1 IF THERE IS A LOWER RESTRICTED
!         RANGE CONSTRAINT AT GRID POINT I, AND INDLW(I)=0
!         OTHERWISE (I=1,...,NUMGR).
!
!  ALTBL  (INPUT) ALTBL(I) CONTAINS THE LOWER BOUND AT GRID POINT I
!         (THUS WE ARE FORCING (P/Q)(GRID POINT I) >= ALTBL(I))
!         IF THERE IS ONE, AND ALTBL(I) NEED NOT BE ASSIGNED A
!         VALUE OTHERWISE (I=1,...,NUMGR).
!
!  OPTION 3  CONSTRAINTS (UPPER RESTRICTED RANGE)
!
!            TO ACTIVATE THIS OPTION, SET THE HUNDREDS DIGIT OF
!            IOPT TO 1 AND SET INDUP AND UPTBL AS FOLLOWS.
!
!  INDUP  (INPUT) INDUP(I)=1 IF THERE IS AN UPPER RESTRICTED
!         RANGE CONSTRAINT AT GRID POINT I, AND INDUP(I)=0
!         OTHERWISE (I=1,...,NUMGR).
!
!  UPTBL  (INPUT) UPTBL(I) CONTAINS THE UPPER BOUND AT GRID POINT I
!         (THUS WE ARE FORCING (P/Q)(GRID POINT I) <= UPTBL(I))
!         IF THERE IS ONE, AND UPTBL(I) NEED NOT BE ASSIGNED A VALUE
!         OTHERWISE (I=1,...,NUMGR).
!
!  OPTION 4  CONSTRAINTS (ADDITIONAL LINEAR CONSTRAINTS)
!
!            TO ACTIVATE THIS OPTION, SET THE ONES DIGIT OF
!            IOPT TO 2 (IF OPTION 1 IS NOT USED) OR TO 3
!            (IF OPTION 1 IS USED) AND SET MEQ, MINEQ, AND
!            ALC AS FOLLOWS.
!
!  MEQ    (INPUT)  THIS IS THE NUMBER OF ADDITIONAL EQUALITY CONSTRAINTS.
!
!  MINEQ  (INPUT)  THIS IS THE NUMBER OF ADDITIOINAL INEQUALITY CONSTRAINTS.
!
!  ALC    (INPUT)  FOR I=1,...,MEQ, THE ITH ROW OF ALC SPECIFIES
!         THE EQUALITY CONSTRAINT
!           ALC(I,1)*X(1)+...+ALC(I,NUMP+NUMQ)*X(NUMP+NUMQ) =
!           ALC(I,NUMP+NUMQ+1)
!         AND FOR I=MEQ+1,...,MEQ+MINEQ, THE ITH ROW OF ALC
!         SPECIFIES THE INEQUALITY CONSTRAINT
!           ALC(I,1)*X(1)+...+ALC(I,NUMP+NUMQ)*X(NUMP+NUMQ) <=
!           ALC(I,NUMP+NUMQ+1).
!
!  OPTION 5  IGNORE F AT SPECIFIED POINTS
!
!            TO ACTIVATE THIS OPTION, SET THE THOUSANDS DIGIT OF
!            IOPT TO 1 AND SET INDF AS FOLLOWS.
!
!  INDF   (INPUT) INDF(I)=1 IF F IS TO BE APPROXIMATED AT GRID
!         POINT I, AND INDF(I)=0 IF F IS TO BE IGNORED AT GRID
!         POINT I (ALTHOUGH OPTIONS 1, 2, AND 3 CAN BE USED AT
!         SUCH POINTS) (I=1,...,NUMGR).  NEITHER FTBLE(I) NOR
!         WTBLE(I) NEED BE ASSIGNED A VALUE IF INDF(I)=0.
!
!            PROGRAMMING HINT... IF OPTION 5 IS USED WITH A NONTRIVIAL
!            DENOMINATOR, IT IS A GOOD IDEA TO ALSO USE OPTION 1 OR 4
!            TO IMPOSE A POSITIVE LOWER BOUND ON THE DENOMINATOR AT
!            LEAST AT THE POINTS WHERE F IS IGNORED, SINCE THE
!            DIFFERENTIAL CORRECTION INQUALITIES ARE NOT PRESENT TO
!            CONTROL THE DENOMINATOR AT SUCH POINTS, AND A DENOMINATOR
!            VALUE WHICH IS SMALL IN ABSOLUTE VALUE OR NEGATIVE COULD
!            TERMINATE THE PROGRAM BEFORE A BEST APPROXIMATION IS FOUND.
!            A REASONABLE LOWER BOUND IS 1000.0*B**(-T) (WHERE T IS THE
!            NUMBER OF BASE B DIGITS IN THE MANTISSA).  NOTE THAT AT
!            POINTS WHERE F IS NOT IGNORED, IF
!            (ABS(F*Q-WT*P)-DELTK*Q)/QK <= W < 0.0, THEN
!            Q >= (ABS(F*Q-WT*P)+ABS(W)*QK)/DELTK > 0.0
!
!  OPTION 6  WEIGHTED APPROXIMATION
!
!            TO ACTIVATE STRAIGHT WEIGHTED APPROXIMATION, THAT IS,
!            TO MINIMIZE MAX(ABS(F-WT*P/Q)), SET THE TEN THOUSANDS
!            DIGIT OF IOPT TO 1 AND SET WTBLE AS BELOW.
!
!            TO ACTIVATE WEIGHTED ERROR CURVE APPROXIMATION, THAT IS,
!            TO MINIMIZE MAX(ABS(WT*(F-P/Q))), SET THE TEN THOUSANDS
!            DIGIT OF IOPT TO 2 AND SET WTBLE AS BELOW.
!
!  WTBLE  (INPUT) WTBLE(I) IS THE VALUE OF THE WEIGHT FUNCTION WT
!         (WHICH NEED NOT BE POSITIVE) AT GRID POINT I (I=1,...,NUMGR).
!
!  OPTION 7  USER SUPPLIED INITIAL APPROXIMATION
!
!            TO ACTIVATE THIS OPTION, SET THE HUNDRED THOUSANDS DIGIT
!            OF IOPT TO 1 AND PLACE THE NUMP+NUMQ INITIAL
!            COEFFICIENTS IN X (WHERE X IS DEFINED IN THE
!            DESCRIPTION SECTION ABOVE).  SOMETIMES ROUND OFF ERROR
!            CAN BE ALLEVIATED, AND COMPUTER TIME CAN BE SAVED, BY
!            FIRST RUNNING A PROBLEM WITH A COARSE GRID AND THEN
!            USING THE RESULT TO INITIALIZE THE PROBLEM ON THE DESIRED GRID.
!
! THE MEANINGS OF SPECIFIC NONZERO VALUES OF IFLAG ARE AS FOLLOWS.
!
!   IFLAG POSITIVE (INITIALIZATION FAILURE) RARELY OCCURS IF THE DATA FOR
!     THE PROGRAM IS ENTERED CORRECTLY.  THERE ARE THREE POSSIBILITIES.
!
!       IFLAG=7777 MEANS OPTION 4 ABOVE WAS USED, AND THE
!         LINEAR INEQUALITY CONSTRAINTS APPEARED TO BE INCONSISTENT.
!
!       IFLAG=3333 MEANS THE APPROXIMATION RETURNED TO THE USER IS
!         THE INITIAL APPROXIMATION P/Q=0.0/PWR(.,NUMP+1) (THAT IS,
!         ALL COEFFICIENTS ARE 0.0 EXCEPT THE FIRST DENOMINATOR
!         COEFFICIENT, WHICH IS 1.0).  THIS APPROXIMATION IS LIKELY
!         TO BE OF LITTLE VALUE.
!
!       IFLAG=6666 MEANS EVEN THE APPROXIMATION P/Q=0.0/PWR(.,NUMP+1)
!         WAS REJECTED BY SUBROUTINE SPQERD BECAUSE ITS DENOMINATOR
!         WAS VERY SMALL IN ABSOLUTE VALUE OR NEGATIVE AT SOME GRID
!         POINT, OR ELSE THE APPROXIMATION VIOLATED SOME CONSTRAINT
!         BY MORE THAN EPRES = 10000.0*B**(-T).  NO APPROXIMATION
!         WAS RETURNED, ALL THE COEFFICIENTS OF P AND Q WERE SET
!         TO 0.0, AND ITER WAS SET TO -1.
!
!   IN CASE IFLAG IS NEGATIVE (INDICATING CAUTION), IT MAY BE
!     ADVISABLE TO EXAMINE THE ERRORS AT THE GRID POINTS, AS
!     NORMALLY THE NUMBER OF POINTS WHERE THE ERROR NORM IS
!     ACHIEVED OR A USER SUPPLIED CONSTRAINT IS SATISFIED WITH
!     EQUALITY, PLUS THE NUMBER OF DENOMINATOR COEFFICIENTS
!     WITH ABSOLUTE VALUE 1.0, WILL BE AT LEAST NUMP+NUMQ+1.
!     THE SPECIFIC POSSIBILITIES FOR IFLAG NEGATIVE, MORE THAN
!     ONE OF WHICH CAN OCCUR, ARE AS FOLLOWS.
!
!       IF THE ONES DIGIT OF IFLAG IS 1, THE INITIAL APPROXIMATION
!         WAS NOT IMPROVED, AND WAS RETURNED TO THE USER.  THIS
!         FREQUENTLY HAPPENS NATURALLY (AND WITH NO ILL
!         CONSEQUENCES) WHEN ONE IS DOING GENERALIZED POLYNOMIAL
!         APPROXIMATION, THAT IS, WHEN ONE HAS SET NUMQ=1 AND
!         PWR(I,NUMP+1)=1.0 FOR I=1,...,NUMGR.
!
!       IF THE ONES DIGIT OF IFLAG IS 2, THE PROGRAM WAS TERMINATED
!         BECAUSE ITLIM ITERATIONS WERE PERFORMED, WHERE CURRENTLY
!         ITLIM=100.  ONE MAY WISH TO RESTART THE PROGRAM,
!         INITIALIZING WITH THE CURRENT APPROXIMATION, TO SEE IF THIS
!         APPROXIMATION CAN BE IMPROVED.
!
!       IF THE TENS DIGIT OF IFLAG IS 1, THEN DURING THE COMPUTATION
!         OF THE RETURNED APPROXIMATION IT WAS NECESSARY TO ADJUST
!         THE TOLERANCES IN SUBROUTINE SLNPRO TO AVOID FAILURE.
!         THUS THE DANGER THAT THE RETURNED APPROXIMATION HAS BEEN
!         SERIOUSLY AFFECTED BY ROUND OFF ERROR IS GREATER.
!
!       IF THE HUNDREDS DIGIT OF IFLAG IS 1, A USER SUPPLIED
!         CONSTRAINT WAS VIOLATED BY MORE THAN EPRES, WHERE
!         EPRES=10000.0*B**(-T).  (B IS THE BASE FOR FLOATING POINT
!         NUMBERS FOR THE MACHINE AND T IS THE NUMBER OF BASE B
!         DIGITS IN THE MANTISSA.)
!
! IF IT IS DESIRED TO CHANGE FROM DOUBLE PRECISION TO SINGLE PRECISION
! (FOR EXAMPLE), ONLY THREE THINGS NEED TO BE DONE.
!
!    (1)  REPLACE ALL THE STATEMENTS  ONE=1.0D0 BY  ONE=1.0
!
!    (2)  REPLACE ALL THE STATEMENTS  IMPLICIT REAL*8 (A-H,O-Z)  BY
!         C       IMPLICIT REAL*8 (A-H,O-Z)
!
!    (3  REPLACE ALL OCCURRENCES OF  D1MACH  BY  R1MACH
!        (OR MODIFY THIS PROGRAM OR THE SUBPROGRAM D1MACH SUPPLIED BY
!        NETLIB SO THAT WHEREVER  THE STATEMENT  SPCMN=D1MACH(3)
!        APPEARS IN THE CURRENT VERSION, SPCMN WILL BE COMPUTED AS
!        B**(-T), WHERE B IS THE BASE FOR FLOATING POINT NUMBERS AND
!        T IS THE NUMBER OF BASE B DIGITS IN THE MANTISSA)
!        (NOTE THAT THE ONLY VALUES WHICH THIS VERSION OF THIS PACKAGE
!        USES FROM THE NETLIB SUPPLIED SUBPROGRAMS I1MACH, R1MACH, AND
!        D1MACH ARE D1MACH(3), I1MACH(1), AND I1MACH(2), WITH THE LATTER
!        TWO VALUES USED ONLY IN THE SAMPLE DRIVER PROGRAM.)
!
! THE PATTERN FOR DIMENSION DECLARATIONS WHICH WILL SET ASIDE
! SUFFICIENT SPACE TO SOLVE ANY PROBLEM WITH UP TO MFN BASIS
! FUNCTIONS (THUS NUMP+NUMQ <= MFN), MGR GRID POINTS (THUS
! NUMGR <= MGR), AND MALC ADDITIONAL LINEAR CONSTRAINTS IN
! OPTION 4 ABOVE (THUS MEQ+MINEQ <= MALC) IS
!
! FTBLE(MGR), PWR(MGR,MFN), INDLW(MGR), ALTBL(MGR), INDUP(MGR),
! UPTBL(MGR), INDF(MGR), WTBLE(MGR), X(MFN+1), ERROR(MGR+1),
! T(2*MGR), ALC(MALC,MFN+1), XTEMP(MFN+1), FTBWT(MGR), IYRCT(MAXCN),
! IYSAV(MAXCN), V(MAXCN+1,MFN+2), IYCCT(MFN+1), IXRCT(MAXCN), Y(MAXCN),
! EQ(MFN,MFN+1), NSCOL(MFN), IRESL(2*MFN),
!
! WHERE MAXCN=5*MGR+2*MFN+MALC-2.  NOTE THAT THE LAST 11 ARRAYS LISTED
! ARE INTERNAL WORK ARRAYS AND NEED NOT BE DIMENSIONED IN THE
! USERS DRIVER PROGRAM.  WE NOTE THAT FOR EACH OF OPTIONS 2 OR 3
! WHICH IS NOT USED, MAXCN COULD BE REDUCED BY MGR, AND MAXCN COULD BE
! REDUCED BY MALC IF OPTION 4 IS NOT USED.
!
! THE METHOD USED BY THIS PROGRAM IS THE DIFFERENTIAL CORRECTION
! ALGORITHM, WHICH REQUIRES AT EACH ITERATION THE SOLUTION OF THE
! LINEAR PROGRAMMING PROBLEM
!     MINIMIZE MAX(ABS(F*Q-WT*P)-DELTK*Q)/QK,
!     SUBJECT TO ABS(Q(J)) <= 1.0 FOR J=1,...,NUMQ,
!       (AND THE OTHER INEQUALITY CONSTRAINTS)
! WHERE QK AND DELTK ARE THE DENOMINATOR AND THE UNIFORM ERROR
! NORM FOR THE PREVIOUS APPROXIMATION, WT IS THE WEIGHT FUNCTION
! (NOTE THAT WT WILL BE 1.0 FOR UNWEIGHTED APPROXIMATION, AND F
! WILL HAVE BEEN MULTIPLIED BY WT FOR WEIGHTED ERROR CURVE
! APPROXIMATION), AND Q(J) IS THE JTH COEFFICIENT OF THE NEW Q.
! HERE MAX MEANS THE MAXIMUM COMPUTED OVER THE GRID.
! THE EQUALITY CONSTRAINTS (IF ANY, SEE OPTION 4 ABOVE) ARE
! HANDLED BY SOLVING THEM FOR MEQA OF THE COEFFICIENTS IN
! TERMS OF THE REMAINING COEFFICIENTS (USING A TOTAL
! PIVOTING STRATEGY) AND IN EFFECT SUBSTITUTING TO REDUCE
! THE NUMBER OF VARIABLES IN THE REST OF THE PROBLEM BY MEQA.
! IN THEORY, THE ERROR NORMS OF THE APPROXIMATIONS PRODUCED BY
! THE ALGORITHM WILL CONVERGE TO THE INFIMUM OF ALL POSSIBLE
! ERROR NORMS.
!
! IN ORDER TO SPEED UP CONVERGENCE FAR FROM THE SOLUTION, FOR
! ITERATIONS BEYOND THE INITIALIZATION AND FIRST MAIN ITERATION
! THE DELTK IN THE ABOVE EXPRESSION IS REPLACED BY 0.5*DELTK
! UNTIL THE PROGRAM FAILS TO PRODUCE A BETTER APPROXIMATION,
! AFTER WHICH THE PROGRAM REVERTS TO ORDINARY DIFFERENTIAL
! CORRECTION FROM THEN ON.  IN THEORY THIS PROCESS, KNOWN AS
! PARAMETRIC BISECTION, WILL RESULT IN MORE THAN HALVING THE
! ERROR NORM AT EACH ITERATION IN WHICH THE ERROR NORM IS
! MORE THAN TWICE THE OPTIMAL ERROR NORM.  THIS IDEA IS DUE
! TO ALI SAZEGARI.
!
! IN THE EVENT OF FAILURE OF THE LINEAR PROGRAMMING SUBROUTINE
! SLNPRO TO PRODUCE AN APPROXIMATION P/Q BETTER THAN THE PREVIOUS
! ONE (IN THE ERROR NORM SENSE), THE PROGRAM TRIES TWICE MORE, ONCE
! WITH LOOSENED AND ONCE WITH TIGHTENED TOLERANCES IN SLNPRO.  AN
! APPROXIMATION PRODUCED BY SLNPRO IS TESTED IN SUBROUTINE SPQERD TO
! SEE IF ITS DENOMINATOR IS NEGATIVE OR VERY SMALL IN ABSOLUTE VALUE
! (THAT IS, LESS THAN 100.0*B**-(T)) AT ANY GRID POINT.  IF SO, THE
! APPROXIMATION IS REJECTED (EXCEPT IN THE INITIALIZATION, WHERE
! THE PROGRAM RESETS THE DENOMINATOR AT THE BAD POINTS AND
! ATTEMPTS TO CONTINUE, AND IF A BETTER APPROXIMATION IS NOT
! PRODUCED, THEN THIS APPROXIMATION IS REJECTED).
! TO SAVE TIME, THE TWO EXTRA LINEAR PROGRAMMING ATTEMPTS ARE
! NOT DONE IF THE FAILURE OCCURS DURING THE PARAMETRIC BISECTION
! PHASE SINCE IN THAT PHASE THE PARAMETRIC BISECTION ATTEMPT WAS
! THE PROBABLE CAUSE OF THE FAILURE, AND THE PROGRAM WILL HAVE
! ANOTHER CHANCE USING ORDINARY DIFFERENTIAL CORRECTION.
!
! THE INITIAL APPROXIMATION FOR THE ALGORITHM IS OBTAINED BY ONE
! OF THREE METHODS, WITH THE USER BEING ABLE TO SELECT THE FIRST
! OR (BY DEFAULT) THE SECOND.  IF A METHOD FAILS, THE PROGRAM
! WILL TRY THE NEXT METHOD IN THE SEQUENCE.  FAILURE OF A METHOD
! MEANS EITHER NO RATIONAL APPROXIMATION P/Q COULD BE PRODUCED,
! OR ONE WAS PRODUCED WHICH VIOLATED SOME CONSTRAINT BY MORE THAN
! EPRES, OR WHICH HAD A DENOMINATOR WHICH WAS VERY SMALL IN
! ABSOLUTE VALUE OR NEGATIVE AT SOME POINT (AND THE MAIN
! ALGORITHM WAS UNABLE TO CORRECT THIS PROBLEM).
! THE THREE METHODS ARE
!
!  METHOD 1  THE USER SUPPLIES AN INITIAL APPROXIMATION (SEE
!            OPTION 7 ABOVE).
!
!  METHOD 2  ONE ITERATION OF LOEBS ALGORITHM IS PERFORMED, THAT
!            IS, SUBROUTINE SLNPRO IS USED TO SOLVE THE PROBLEM
!                MINIMIZE MAX(ABS(F*Q-WT*P))
!                SUBJECT TO Q >= MAX(0.00001,10000.0*B**(-T))
!                EVERYWHERE ON THE GRID, ABS(Q(J)) <= 1.0
!                FOR J=1,...,NUMQ (AND THE OTHER
!                INEQUALITY CONSTRAINTS).
!            IF WE ARE PLACING A LOWER BOUND ON THE DENOMINATOR
!            IN THE MAIN ITERATIONS, THE LOWER BOUND HERE IS
!            INCREASED (IF NECESSARY) TO MATCH THAT BOUND.
!            IF SLNPRO FAILS TO PRODUCE AN APPROXIMATION P/Q WE
!            TRY WITH BOTH LOOSENED AND TIGHTENED TOLERANCES.
!
!  METHOD 3  THE INITIAL APPROXIMATION IS TAKEN TO BE P/Q=
!            0.0/PWR(.,NUMP+1).
!
! THE ALGORITHM IS TERMINATED WHEN A BETTER APPROXIMATION CANNOT
! BE PRODUCED OR ITLIM ITERATIONS HAVE BEEN PERFORMED, AND THE
! BEST APPROXIMATON FOUND SO FAR IS THEN RETURNED TO THE USER.
! IF IT WAS NECESSARY TO ADJUST THE TOLERANCES IN SLNPRO TO
! OBTAIN AN APPROXIMAITON P/Q, THEN VIOLATION OF A USER SUPPLIED
! CONSTRAINT BY MORE THAN 10000.0*B**(-T) WILL CAUSE THAT
! APPROXIMATION TO BE REJECTED SINCE THE ADJUSTED TOLERANCES
! MAY HAVE ALLOWED REDUCTION OF THE UNIFORM ERROR NORM AT THE
! EXPENSE OF A SERIOUS VIOLATION OF THE USER SUPPLIED CONSTRAINTS.
!
!***REFERENCES  CHENEY, E. W. AND LOEB, H. L., TWO NEW ALGORITHMS FOR
!                 RATIONAL APPROXIMATION, NUMER. MATH. 4, PP. 124-127, 1962.
!               BARRODALE, I., POWELL, M. J. D., AND ROBERTS, F. D. K.,
!                 THE DIFFERENTIAL CORRECTION ALGORITHM FOR RATIONAL
!                 L-INFINITY APPROXIMATION, SIAM J. NUMER. ANAL. 9,
!                 PP. 493-504, 1972.
!               KAUFMAN, EDWIN H. JR., LEEMING, D. J., AND TAYLOR, G. D.,
!                 UNIFORM RATIONAL APPROXIMATION BY DIFFERENTIAL CORRECTION
!                 AND REMES-DIFFERENTIAL CORRECTION, INT. J. FOR NUMER.
!                 METH. IN ENGRG. 17, PP. 1273-1278, 1981.
!***REFER TO  (NONE)
!***ROUTINES CALLED  SLNPRO,SPQERD,SSETV,EQELM,SCRCH,EXPND,
!                    ICNCK,NRMLZ
!***END PROLOGUE  CDFCOR
!

INTEGER, INTENT(IN)        :: nump
INTEGER, INTENT(IN)        :: numq
INTEGER, INTENT(IN)        :: numgr
REAL (dp), INTENT(IN)      :: ftble(101)
REAL (dp), INTENT(IN OUT)  :: pwr(101,15)
INTEGER, INTENT(IN)        :: iopt
REAL (dp), INTENT(IN OUT)  :: epsq
INTEGER, INTENT(IN OUT)    :: indlw(101)
REAL (dp), INTENT(IN OUT)  :: altbl(101)
INTEGER, INTENT(IN OUT)    :: indup(101)
REAL (dp), INTENT(IN OUT)  :: uptbl(101)
INTEGER, INTENT(IN OUT)    :: indf(101)
REAL (dp), INTENT(IN OUT)  :: wtble(101)
INTEGER, INTENT(IN OUT)    :: meq
INTEGER, INTENT(IN OUT)    :: mineq
REAL (dp), INTENT(IN OUT)  :: alc(101,16)
INTEGER, INTENT(IN OUT)    :: meqa
REAL (dp), INTENT(OUT)     :: x(16)
REAL (dp), INTENT(OUT)     :: error(102)
REAL (dp), INTENT(OUT)     :: qmin
INTEGER, INTENT(OUT)       :: iqmin
INTEGER, INTENT(OUT)       :: iter
INTEGER, INTENT(OUT)       :: indlp
INTEGER, INTENT(OUT)       :: iflag

!
REAL (dp)  :: v(635,17), xtemp(16), ftbwt(101), EQ(15,16)
INTEGER    :: iyrct(634), iysav(634), nscol(15), iresl(30)
REAL (dp)  :: deltk, dlold, epres, four, one, spcmn, ten, tenth, tolel, &
              two, zero
INTEGER    :: i, iadjs, icon, icon0, icon1, ilbfl, ilst1, indic, initl, &
              ione, ipbit, iret1, itlim, j, m, n, nlp, nmpp1, nmppq, np1
!
! SET MACHINE DEPENDENT PARAMETERS FOR SUBROUTINE CDFCOR.
! SET SPCMN=B**(-T), WHERE B IS THE BASE AND T IS THE NUMBER
! OF BASE B DIGITS IN THE MANTISSA.  SPCMN IS THE MINIMUM
! RELATIVE SPACING ABS((X1-X2)/X2) BETWEEN TWO SUCCESSIVE
! FLOATING POINT NUMBERS, SO IT IS THE SPACING BETWEEN TWO
! SUCCESSIVE FLOATING POINT NUMBERS IN THE CLOSED INTERVAL
! (0.1,1.0).  WE ALSO HAVE SPCMN=10.0**(-T*(LOG10)(B))=
! 10.0**(-TNMAN), WHERE TNMAN IS THE BASE 10 EQUIVALENT OF
! THE LENGTH OF THE MANTISSA.
!
!***FIRST EXECUTABLE STATEMENT  CDFCOR
spcmn=EPSILON(1.0_dp)

! SET PRECISION DEPENDENT CONSTANTS FOR CDFCOR.
one=1.0_dp
zero=0.0_dp
two=one+one
four=two+two
ten=four+four+two
tenth=one/ten
! SET EPRES=10000.0*SPCMN.  THUS EPRES=10.0**(-(TNMAN-4)).
! AN APPROXIMATION WILL BE REJECTED IF A USER SUPPLIED
! CONSTRAINT IS VIOLATED BY MORE THAN EPRES AND IT WAS NECESSARY
! TO ADJUST THE TOLERANCES IN SUBROUTINE SLNPRO WHILE COMPUTING
! THE APPROXIMATION.
epres=spcmn*ten**4
! SET TOLEL=100.0*SPCMN.  IN SUBROUTINE EQELM, NUMBERS WITH
! ABSOLUTE VALUE <= TOLEL WILL BE TREATED AS ZEROES.
! THIS WILL ALSO BE THE CASE IN SUBROUTINE NRMLZ.
tolel=ten*ten*spcmn
! END OF SETTING MACHINE AND PRECISION DEPENDENT PARAMETERS FOR CDFCOR.
!
itlim=100
iter=0
nmpp1=nump+1
nmppq=nump+numq
n=nmppq+1
np1=n+1
nlp=n
iadjs=0
indic=0
iret1=0
ilbfl=0
ilst1=0
qmin=zero
iqmin=0
initl=0
iflag=0
ipbit=1
ione=iopt - (iopt/10)*10

! Initialize V
v = zero

! IF THE ONES DIGIT OF IOPT IS <= 1, SET MEQ AND MINEQ TO ZERO.
IF (ione <= 1) THEN
  meq=0
  mineq=0
END IF
meqa=0
!
! IF F IS NOT TO BE IGNORED AT ANY OF THE GRID POINTS,
! SET INDF IDENTICALLY EQUAL TO 1.
IF (iopt-(iopt/10000)*10000 < 1000) indf(1:numgr)=1
!
! IF WE ARE DEALING WITH UNWEIGHTED APPROXIMATION SET THE
! WEIGHT FUNCTION IDENTICALLY EQUAL TO 1.
IF (iopt-(iopt/100000)*100000 < 10000) wtble(1:numgr)=one
!
! COPY FTBLE INTO FTBWT FOR USE BY THE REST OF THIS PROGRAM
! UNLESS WEIGHTED ERROR CURVE APPROXIMATION IS DESIRED.  IN
! THAT CASE SET FTBWT(I)=WTBLE(I)*FTBLE(I) AT THOSE POINTS
! WHERE F IS TO BE APPROXIMATED, SO WE WILL HAVE
! FTBWT(I)-WTBLE(I)*P/Q=WTBLE(I)*(FTBLE(I)-P/Q).  THE VECTOR
! FTBWT IS INTRODUCED SO THAT FTBLE NEED NOT BE CHANGED BY
! THE PROGRAM.
IF (iopt-(iopt/100000)*100000 < 20000) THEN
  DO  i=1,numgr
    IF (indf(i) > 0) ftbwt(i)=ftble(i)
  END DO

ELSE
  DO  i=1,numgr
    IF (indf(i) > 0) ftbwt(i)=wtble(i)*ftble(i)
  END DO

END IF

!
! IF THERE ARE EQUALITY CONSTRAINTS, ATTEMPT TO SOLVE FOR MEQA OF THE
! COEFFICIENTS IN TERMS OF THE REMAINING NUMP+NUMQ-MEQA COEFFICIENTS,
! WHERE MEQA IS THE NUMBER OF INDEPENDENT EQUALITY CONSTRAINTS.
IF (meq > 0) THEN

! HERE THERE ARE MEQ EQUALITY CONSTRAINTS, AND WE COPY THE FIRST
! MEQ ROWS OF ALC INTO EQ.
  EQ(1:meq,1:n) = alc(1:meq,1:n)
  CALL eqelm (nmppq, meq, tolel, EQ, nscol, iresl, meqa, iflag)

! HERE IFLAG WILL BE EITHER 0 OR 7777.  IN THE LATTER CASE THE
! EQUALITY CONSTRAINTS WERE INCONSISTENT AND WE RETURN.
  IF (iflag > 0) GO TO 960

! RESET N AND NP1 FOR USE IN SLNPRO.
  nlp=n-meqa
END IF
!
! IF THE 100000 DIGIT OF IOPT IS 1, THEN THE
! INITIALIZATION BY LOEBS ALGORITHM WILL BE SKIPPED, AND
! THE COEFFICIENTS PLACED IN X BY THE USER WILL BE TAKEN FOR
! THE INITIAL APPROXIMATION.
IF (iopt/100000 > 0) THEN

! SET INITL=-1 TO INDICATE WE ARE USING A USER SUPPLIED INITIAL APPROXIMATION.
  initl=-1
  GO TO 340
END IF
!
! HERE WE ATTEMPT TO COMPUTE AN INITIAL APPROXIMATION USING LOEBS ALGORITHM.
220 initl=0

! WE TRY TO FIND P, Q SUCH THAT Q IS POSITIVE (AND NOT TOO
! SMALL) AT EVERY GRID POINT, P AND Q SATISFY THE CONSTRAINTS
! (INCLUDING THE CONSTRAINTS MAX(ABS(Q(J))) <= 1.0), AND
! MAX(ABS(F*Q-WT*P)) IS MINIMIZED AT THOSE GRID POINTS WHERE
! F IS TO BE APPROXIMATED.
230 CALL ssetv (nump, numq, numgr, ftbwt, pwr, iopt, epsq, indlw,  &
                altbl, indup, uptbl, indf, wtble, meq, mineq, alc, 0, v, m)
! SET IYRCT(1)=-1 AS A SIGNAL TO DO THE INITIAL EXCHANGES
! IN SLNPRO STRICTLY ACCORDING TO A PIVOTING STRATEGY.
iyrct(1)=-1

! REDUCE V IF MEQA > 0
IF (meqa > 0) CALL scrch (m, nmppq, meqa, nscol, iresl, EQ, v)

! IF SLNPRO FAILED DURING THE INITIALIZATION WITH NORMAL TOLERANCES,
! AS INDICATED BY IADJS=1, WE SET M=-M AS A SIGNAL TO LOOSEN THE
! TOLERANCES IN SLNPRO (I. E. INCREASE REA AND REA1) AND TRY AGAIN.
! SLNPRO WILL RESET M TO ITS ORIGINAL VALUE.  ANY LOSS OF ACCURACY
! DUE TO THIS LOOSENING WILL PROBABLY BE CORRECTED IN LATER ITERATIONS.
m=m*(1 - 4*iadjs + 2*iadjs**2)

! IF SLNPRO FAILED DURING THE INITIALIZATION WITH BOTH NORMAL
! TOLERANCES AND LOOSENED TOLERANCES, AS INDICATED BY IADJS=2,
! WE SET NLP=-NLP AS A SIGNAL TO TIGHTEN THE TOLERANCES IN SLNPRO
! (I. E. DECREASE REA AND REA1) AND TRY ONCE AGAIN.  SLNPRO WILL
! RESET NLP TO ITS ORIGINAL VALUE.
nlp=nlp*(1+iadjs-iadjs**2)
CALL slnpro (v, m, nlp, iyrct, x, indic)

! SAVE INDIC IN CASE THE APPROXIMATION JUST COMPUTED IS THE ONE RETURNED.
iret1=indic

! IF THIS WAS THE FIRST LOEB ATTEMPT WE SAVE INDIC IN ILBFL FOR
! OUTPUT PURPOSES IN CASE BOTH LOEB AND THE FIRST MAIN ITERATION FAIL.
IF (iadjs <= 0) ilbfl=indic

! IF INDIC IS POSITIVE SOMETHING WENT WRONG IN SLNPRO.
IF (indic <= 0) GO TO 320
IF (iadjs <= 1) THEN

! BEFORE CONCEDING DEFEAT WE WILL TRY AGAIN WITH CHANGED TOLERANCES.
  iadjs=iadjs+1
  GO TO 230
END IF
!
! HERE THE LOEB INITIALIZATION HAS FAILED, AND WE TAKE OUR INITIAL
! APPROXIMATION TO BE P/Q=0.0/PWR(.,NUMP+1).  WE ALSO SET INITL=1
! AS A WARNING THAT THIS HAS BEEN DONE.
300 initl=1
DO  j=1,nmppq
  x(j)=zero
END DO
x(nmpp1)=one
GO TO 350
!
! HERE INDIC WAS NONPOSITIVE, SO SLNPRO PRODUCED AN APPROXIMATION.
! IF MEQA > 0, EXPAND X TO INCLUDE THE MEQA VARIABLES
! SOLVED FOR IN SUBROUTINE EQELM.
320 IF (meqa > 0) THEN
  CALL expnd (nmppq, meqa, nscol, iresl, EQ, x)
END IF

! NORMALIZE P AND Q IF POSSIBLE.
340 CALL nrmlz (nump, numq, meq, mineq, nscol, iresl, alc, tolel, x)
!
! COMPUTE THE VALUES OF P AND Q AT THE GRID POINTS, THE
! ERRORS AT THE GRID POINTS, AND THE UNIFORM ERROR NORM.
350 CALL spqerd (nump, numq, numgr, ftbwt, pwr, indf, wtble, x, v, qmin, iqmin)
deltk=v(2*numgr+1,np1)
dlold=deltk

! RESET IADJS.
iadjs=0
! PUT THE ERRORS AND ERROR NORM INTO ERROR.
error(1:numgr)=v(1:numgr,n)
error(numgr+1)=deltk
!
! IF THE DENOMINATOR WAS VERY SMALL IN ABSOLUTE VALUE OR
! NEGATIVE AT SOME POINT (WHICH IS INDICATED BY
! V(2*NUMGR+1,NP1) BEING NEGATIVE), WE RESET
! V(2*NUMGR+1,NP1) TO THE ERROR NORM COMPUTED WITH THE BAD
! POINTS IGNORED FOR USE IN SSETV2, AND ATTEMPT TO CONTINUE.
! NOTE THAT DELTK, DLOLD, AND ERROR(NUMGR+1) WILL REMAIN
! NEGATIVE AS A WARNING UNLESS AN ACCEPTABLE APPROXIMATION WITH
! SIGNIFICANTLY POSITIVE DENOMINATOR IS FOUND BELOW.
! THIS SITUATION SOMETIMES OCCURS WHEN AN APPROXIMATION
! COMPUTED ON A COARSE GRID IS USED AS A STARTING
! APPROXIMATION ON A FINE GRID.
IF (deltk < 0.0) THEN
  v(2*numgr+1,np1)=-v(2*numgr+1,np1)-two
! SET ICON0=1 TO INDICATE THIS IS NOT AN ACCEPTABLE APPROXIMATION.
  icon0=1
ELSE
!
! HERE DELTK >= 0.0.  MAKE ICON0=0 IF THE CONSTRAINTS ARE
! SATISFIED WITHIN EPRES AND ICON0=1 OTHERWISE.
  icon0=icnck(iopt, nump, numq, numgr, indlw, altbl, indup, uptbl, epsq, &
              x, v, meq, mineq, alc, epres)
END IF
!
! END OF INITIALIZATION PHASE.
!
!
! THE NEXT GROUP OF STATEMENTS (THROUGH STATEMENT 5400) ARE
! FOR THE FIRST MAIN DIFFERENTIAL CORRECTION ITERATION.
! IF THIS ITERATION FAILS TO PRODUCE AN ACCEPTABLE
! APPROXIMATION AND THE INITIAL APPROXIMATION IS ALSO
! UNACCEPTABLE WE QUIT, OTHERWISE WE TRY THE NEXT
! INITIALIZATION STAGE.
!
! SET UP THE LINEAR PROGRAMMING PROBLEM.
390 CALL ssetv (nump, numq, numgr, ftbwt, pwr, iopt, epsq, indlw,  &
                altbl, indup, uptbl, indf, wtble, meq, mineq, alc, 1, v, m)

! SET IYRCT(1)=-1 AS A SIGNAL TO DO THE INITIAL EXCHANGES IN
! SLNPRO STRICTLY ACCORDING TO A PIVOTING STRATEGY RATHER
! THAN ATTEMPTING TO USE INFORMATION ABOUT A PREVIOUS VERTEX.
iyrct(1)=-1

! REDUCE V IF MEQA > 0
IF (meqa > 0) CALL scrch (m, nmppq, meqa, nscol, iresl, EQ, v)

! SET M=-M AS A SIGNAL TO LOOSEN THE TOLERANCES IN SLNPRO IF
! IADJS=2 AND LEAVE M UNCHANGED IF IADJS=0 OR 1.  SLNPRO WILL
! RESET M TO ITS ORIGINAL VALUE IF IT IS CHANGED HERE.
m=m*(1+iadjs-iadjs**2)

! SET NLP=-NLP AS A SIGNAL TO TIGHTEN THE TOLERANCES IN SLNPRO IF
! IADJS=1 AND LEAVE NLP UNCHANGED IF IADJS=0 OR 2.  SLNPRO WILL
! RESET NLP TO ITS ORIGINAL VALUE IF IT IS CHANGED HERE.
nlp=nlp*(1-4*iadjs+2*iadjs**2)
CALL slnpro (v, m, nlp, iyrct, xtemp, indic)

! IF THIS WAS THE FIRST SLNPRO CALL IN THIS ITERATION, SO THE
! TOLERANCES IN SLNPRO WERE NOT ADJUSTED BY CDFCOR, FOR
! POSSIBLE OUTPUT PURPOSES WE SAVE INDIC IN ILST1.
IF (iadjs <= 0) ilst1=indic

! IF INDIC IS POSITIVE SOMETHING WENT WRONG IN SLNPRO SO WE
! TRY AGAIN WITH BOTH TIGHTER AND LOOSER TOLERANCES IF THIS
! HAS NOT ALREADY BEEN TRIED.
IF (indic <= 0) GO TO 490
!
! HERE WE DO NOT HAVE AN ACCEPTABLE APPROXIMATION.
440 IF (iadjs <= 1) THEN
  iadjs=iadjs+1

! CALL SPQERD WITH THE COEFFICIENTS FROM THE INITIALIZATION
! (WHICH ARE IN X) TO RECOMPUTE THE VALUES OF P AND Q AND
! PUT THEM IN V FOR USE BY SSETV.
  CALL spqerd (nump, numq, numgr, ftbwt, pwr, indf, wtble, x, v, qmin, iqmin)
  GO TO 390
END IF
!
! HERE WE HAVE TRIED AND FAILED WITH NORMAL, TIGHTENED, AND
! LOOSENED TOLEARANCES SO THE FIRST MAIN DIFFERENTIAL
! CORRECTION ITERATION HAS FAILED TO PRODUCE AN ACCEPTABLE
! APPROXIMATION.  IF THE APPROXIMATION FROM THE INITIALIZATION
! PHASE WAS ACCEPTABLE, WE SET IFLAG AND INDLP, AND RETURN.
IF (icon0 <= 0) GO TO 850

! HERE THE INITIAL APPROXIMATION WAS UNACCEPTABLE.  IF IT WAS
! USER SUPPLIED WE RESET IADJS AND TRY LOEB, IF IT CAME
! FROM LOEB WE TRY P/Q = 0.0/1.0, AND IF IT WAS P/Q = 0.0/1.0
! WE SET IFLAG AND INDLP, AND RETURN.
IF (initl < 0) THEN
  GO TO 480
ELSE IF (initl == 0) THEN
  GO TO 300
ELSE
  GO TO 850
END IF

480 iadjs=0
GO TO 220

! END OF BLOCK WITHIN FIRST MAIN DIFFERENTIAL CORRECTION
! ITERATION TO DEAL WITH AN UNACCEPTABLE APPROXIMATION.
!
! HERE INDIC WAS NONPOSITIVE, SO SLNPRO PLACED AN APPROXIMATION
! IN XTEMP, WHICH WE NOW CHECK FOR ACCEPTABLENESS.  FIRST
! EXPAND XTEMP IF MEQA > O
490 IF (meqa > 0) CALL expnd (nmppq, meqa, nscol, iresl, EQ, xtemp)

! CALL SPQERD TO COMPUTE P, Q, THE ERRORS, AND THE
! UNIFORM ERROR NORM (AND TO PLACE THESE VALUES IN V FOR
! USE BY OTHER ROUTINES).
CALL spqerd (nump, numq, numgr, ftbwt, pwr, indf, wtble, xtemp, v, qmin, iqmin)
deltk=v(2*numgr+1,np1)

! IF DELTK IS NEGATIVE (INDICATING A VERY SMALL OR NEGATIVE
! DENOMINATOR VALUE AT SOME POINT) WE REJECT THIS
! APPROXIMATION AND TRY AGAIN.
IF (deltk < 0.0) GO TO 440

! HERE DELTK IS NONNEGATIVE, AND WE NORMALIZE XTEMP AND CHECK THE CONSTRAINTS.
CALL nrmlz (nump, numq, meq, mineq, nscol, iresl, alc, tolel, xtemp)
icon1=icnck(iopt, nump, numq, numgr, indlw, altbl, indup, uptbl, epsq,  &
            xtemp, v, meq, mineq, alc, epres)
! IF THE APPROXIMATION IS UNACCEPTABLE WE TRY AGAIN.
IF (icon1 > 0) GO TO 440

! HERE THE APPROXIMATION COMPUTED BY THE FIRST MAIN CDFCOR
! ITERATION SATISFIES THE CONSTRAINTS WITHIN EPRES, AND WE WILL
! ACCEPT IT IF THE APPROXIMATION FROM THE INITIALIZATION WAS
! UNACCEPTABLE OR DOES NOT HAVE A SMALLER ERROR NORM.
! OTHERWISE WE WILL REJECT IT AND TRY AGAIN SINCE SOMETHING IS FISHY.
IF (icon0 <= 0) THEN
  IF (deltk-dlold > 0.0) GO TO 440
END IF
!
! END OF GROUP OF STATEMENTS FOR FIRST MAIN DIFFERENTIAL
! CORRECTION ITERATION.  FROM NOW ON ONLY STATEMENTS BELOW
! THIS POINT IN CDFCOR WILL BE EXECUTED.
!
!
! HERE THE PRESENT APPROXIMATION WILL BE KEPT, AND WE PREPARE
! TO COMPUTE ANOTHER ONE.
! COPY XTEMP INTO X.
550 DO  i=1,nmppq
  x(i)=xtemp(i)
END DO

! PUT THE ERRORS AND ERROR NORM INTO ERROR.
error(1:numgr)=v(1:numgr,n)
error(numgr+1)=deltk
iter=iter+1
IF (iter < itlim) THEN
  dlold=deltk

! SAVE INDIC IN CASE THE APPROXIMATION JUST COMPUTED IS THE ONE RETURNED.
  iret1=indic
! RESET IADJS.
  iadjs=0

! IF IPBIT=1 WE ARE STILL IN THE PARAMETRIC BISECTION PHASE,
! AND WE CUT V(2*NUMGR+1,N+1) IN HALF IN ORDER TO TRY FOR AT
! LEAST A HALVING OF THE ERROR NORM IN THE NEXT STEP.
  IF (ipbit <= 0) GO TO 630
    v(2*numgr+1,np1)=v(2*numgr+1,np1)/two
ELSE
!
! HERE ITLIM ITERATIONS HAVE BEEN PERFORMED AND WE WILL RETURN
! AFTER SETTING IFLAG AND INDLP (SEE BELOW).
  indlp=5-indic
  IF (indic < 0) THEN
    iflag=-12
    GO TO 940
  END IF
  iflag=-2
  GO TO 940
END IF
!
! SAVE IYRCT IN IYSAV SO WE CAN LATER USE INFORMATION FROM THIS
! SUCCESSFUL SLNPRO ITERATION TO RESTART SLNPRO.
630 DO  i=1,m
  iysav(i)=iyrct(i)
END DO
! CALL SSETV TO SET UP FOR THE NEXT ITERATION.
650 CALL ssetv (nump, numq, numgr, ftbwt, pwr, iopt, epsq, indlw,  &
                altbl, indup, uptbl, indf, wtble, meq, mineq, alc, 1, v, m)
! REDUCE V IF MEQA > 0
IF (meqa > 0) CALL scrch (m, nmppq, meqa, nscol, iresl, EQ, v)

! IF IADJS=1, SET NLP=-NLP (AND LEAVE M ALONE) AS A SIGNAL TO
! TIGHTEN THE TOLERANCES IN SLNPRO, AND IF IADJS=2 SET M=-M
! (AND LEAVE NLP ALONE) AS A SIGNAL TO LOOSEN THE TOLERANCES
! IN SLNPRO.  IF IADJS=0, LEAVE THE TOLERANCES ALONE.  M AND
! NLP WILL BE RESET TO THEIR ORIGINAL VALUES BY SLNPRO IF THEY
! ARE CHANGED HERE.
m=m*(1+iadjs-iadjs**2)
nlp=nlp*(1-4*iadjs+2*iadjs**2)
CALL slnpro (v, m, nlp, iyrct, xtemp, indic)

! IF THIS WAS THE FIRST SLNPRO CALL IN THIS ITERATION, SO THE
! TOLERANCES IN SLNPRO WERE NOT ADJUSTED BY CDFCOR, FOR
! POSSIBLE OUTPUT PURPOSES WE SAVE INDIC IN ILST1.
IF (iadjs <= 0) ilst1=indic

! IF INDIC IS POSITIVE SOMETHING WENT WRONG IN SLNPRO SO WE GO
! INTO THE BLOCK OF STATEMENTS DESIGNED TO HANDLE FAILURE TO
! PRODUCE AN ACCEPTABLE APPROXIMATION.
IF (indic <= 0) GO TO 760
!
! HERE WE DID NOT GET AN ACCEPTABLE APPROXIMATION.  IF WE ARE
! STILL DOING PARAMETRIC BISECTION, TO SAVE TIME WE DO NOT TRY
! SLNPRO AGAIN WITH ADJUSTED TOLERANCES SINCE THE PARAMETRIC
! BISECTION PROBABLY CAUSED THE FAILURE, BUT RATHER WE SWITCH
! PERMANENTLY INTO ORDINARY DIFFERENTIAL CORRECTION.
700 IF (ipbit > 0) THEN
  ipbit=0
ELSE

! IF WE WERE NOT DOING PARAMETRIC BISECTION WE TRY AGAIN WITH
! TIGHTER AND LOOSER TOLERANCES, IF THIS HAS NOT ALREADY BEEN
! TRIED.  IF IT HAS BEEN TRIED, WE SET IFLAG AND INDLP AND RETURN.
  IF (iadjs > 1) GO TO 850
  iadjs=iadjs+1
END IF

! CALL SPQERD WITH THE COEFFICIENTS FROM THE PREVIOUS ITERATION
! (WHICH ARE IN X) TO RECOMPUTE THE VALUES OF P AND Q AND PUT
! THEM IN V FOR USE BY SSETV.
CALL spqerd (nump, numq, numgr, ftbwt, pwr, indf, wtble, x, v, qmin, iqmin)

! RESTORE IYRCT TO ITS VALUE AT THE PREVIOUS ITERATION.
iyrct(1:m)=iysav(1:m)
GO TO 650

! END OF BLOCK OF STATEMENTS WITHIN THE SECOND OR LATER MAIN
! DIFFERENTIAL CORRECTION ITERATION TO HANDLE FAILURE TO
! PRODUCE AN ACCEPTABLE APPROXIMATION.
!
! HERE INDIC WAS NONPOSITIVE, SO SLNPRO PRODUCED AN
! APPROXIMATION IN XTEMP WHICH WE NOW CHECK FOR
! ACCEPTABILITY.  FIRST EXPAND XTEMP IF MEQA > 0
760 IF (meqa > 0) CALL expnd (nmppq, meqa, nscol, iresl, EQ, xtemp)

! CALL SPQERD TO COMPUTE P, Q, THE ERRORS, AND THE UNIFORM
! ERROR NORM (AND PLACE THEM IN V FOR USE BY OTHER ROUTINES).
CALL spqerd (nump, numq, numgr, ftbwt, pwr, indf, wtble, xtemp, v, qmin, iqmin)
deltk=v(2*numgr+1,np1)

! IF DELTK IS NEGATIVE (INDICATING A VERY SMALL OR NEGATIVE
! DENOMINATOR VALUE AT SOME POINT) WE REJECT THIS
! APPROXIMATION AND TRY AGAIN.
IF (deltk < 0.0) GO TO 700

! IF INDIC IS NEGATIVE HERE IT WAS NECESSARY
! TO ADJUST TOLERANCES IN SLNPRO, SO THE APPROXIMATION
! COMPUTED IS MORE LIKELY TO HAVE BEEN AFFECTED
! BY ROUNDOFF ERROR.  THUS IN THIS CASE WE CHECK THE
! CONSTRAINTS, AND REJECT THE APPROXIMATION IF ANY OF THEM ARE
! VIOLATED BY MORE THAN EPRES.  NOTE THAT INDIC CANNOT BE
! POSITIVE HERE SINCE SLNPRO PRODUCED AN APPROXIMATION.
IF (indic < 0) THEN
  icon=icnck(iopt, nump, numq, numgr, indlw, altbl, indup, uptbl, epsq,  &
             xtemp, v, meq, mineq, alc, epres)
  IF (icon > 0) GO TO 700
END IF

! REJECT THIS APPROXIMATION AND TRY AGAIN IF THE ERROR NORM HAS NOT DECREASED.
IF (deltk >= dlold) GO TO 700

! RARELY AN APPROXIMATION WITH ALL COEFFICIENTS VERY SMALL
! WILL BE PRODUCED.  TO BE SAFE, IF ALL DENOMINATOR
! COEFFICIENTS ARE <= 0.1 IN ABSOLUTE VALUE WE ATTEMPT TO
! NORMALIZE THE APPROXIMATION AND CHECK THE CONSTRAINTS.  IF
! THEY ARE SATISFIED WITHIN EPRES WE ACCEPT THIS
! APPROXIMATION, AND OTHERWISE WE TRY AGAIN.
DO  j=nmpp1,nmppq
  IF (xtemp(j) > tenth) GO TO 550
  IF (xtemp(j) < -tenth) GO TO 550
END DO
CALL nrmlz (nump, numq, meq, mineq, nscol, iresl, alc, tolel, xtemp)
icon=icnck(iopt, nump, numq, numgr, indlw, altbl, indup, uptbl, epsq,  &
           xtemp, v, meq, mineq, alc, epres)
IF (icon > 0) THEN
  GO TO 700
ELSE
  GO TO 550
END IF
! END OF COMPUTATION OF APPROXIMATIONS.
!
!
! HERE WE ARE FINISHED COMPUTING APPROXIMATIONS, SO WE SET INDLP,
! IFLAG, QMIN, AND IQMIN, AND WE RETURN.
! WE CARRY INFORMATION BACK TO THE CALLING PROGRAM ABOUT TWO
! SLNPRO CALLS BY SETTING THE TWO DIGIT VARIABLE INDLP WITH
! TENS DIGIT 5-ILAST AND ONES DIGIT 5-IRET, WHERE IN MOST
! CASES ILAST AND IRET GIVE INFORMATION ON THE LAST (REJECTED)
! SLNPRO ATTEMPT AND THE RETURNED APPROXIMATION RESPECTIVELY.
! TO BE EXACT, IF THE PROGRAM WAS TERMINATED BECAUSE ITLIM
! ITERATIONS WERE PERFORMED SUCCESSFULLY, ILAST WILL BE 5 (AND
! SO 5-ILAST WILL BE 0).  OTHERWISE, ILAST WILL BE THE VALUE OF
! INDIC PRODUCED BY SLNPRO AT THE LAST (FAILED) MAIN ITERATION,
! WITH THE TOLERANCES IN SLNPRO NOT HAVING BEEN ADJUSTED BY CDFCOR YET.
! IF THE PROGRAM FAILED IN THE INITIALIZATION (SO IFLAG WAS
! POSITIVE), IRET WILL BE THE VALUE OF INDIC PRODUCED BY
! SLNPRO DURING THE (FAILED) LOEB INITIALIZATION, WITH THE
! TOLERANCES IN SLNPRO NOT HAVING BEEN ADJUSTED BY CDFCOR
! YET.  IF A USER SUPPLIED INITIAL APPROXIMATION WAS RETURNED
! TO THE USER, IRET WILL BE O.  OTHERWISE, IRET WILL BE THE
! VALUE OF INDIC PRODUCED BY SLNPRO DURING THE COMPUTATION
! OF THE APPROXIMATION WHICH WAS ACTUALLY RETURNED TO THE USER.
! THUS IF THE APPROXIMATION RETURNED WAS COMPUTED BY SLNPRO
! (WHICH WILL BE THE CASE IF EITHER THE ONES DIGIT OF IFLAG IS
! 0 OR 2, OR IT IS 1 AND THE RETURNED APPROXIMATION IS NOT A USER
! SUPPLIED INITIAL APPROXIMATION) AND THE ONES DIGIT OF INDLP IS 5,
! THEN SLNPRO APPEARED TO WORK NORMALLY DURING THE COMPUTATION OF
! THE RETURNED APPROXIMATION.
850 indlp=10*(5-ilst1)+5-iret1
!
! WE NOW SET UP IFLAG FOR THE INFORMATION OF THE USER.
IF (iter <= 0) THEN
!
! HERE ITER=0, SO THE INITIAL APPROXIMATION WAS NOT IMPROVED.
  IF (initl > 0) THEN

! HERE THE UNIMPROVED INITIAL APPROXIMATION WAS P/Q=
! 0.0/PWR(.,NUMP+1).  IF IT WAS FOUND BY SPQERD TO HAVE A
! DENOMINATOR WHICH IS VERY SMALL IN ABSOLUTE VALUE OR
! NEGATIVE AT SOME POINT, OR IF IT VIOLATED A CONSTRAINT (EITHER OF
! WHICH WOULD BE INDICATED BY ICON0=1), THEN SET IFLAG=6666,
! ITER=-1, AND ZERO OUT THE ERRORS AND COEFFICIENTS AND RETURN.
! OTHERWISE SET IFLAG=3333 AND RETURN.
! FIRST RESET INDLP SO THAT 5-(ONES DIGIT OF INDIC) IS THE VALUE
! OF INDIC THE FIRST TIME SLNPRO WAS CALLED IN THE (FAILED) LOEB
! INITIALIZATION.
    indlp=10*(5-ilst1) + 5 - ilbfl
    IF (icon0 > 0) THEN
      iflag=6666
      iter=-1
! NOTE THAT HERE X(I) ALREADY EQUALS 0.0 FOR I=1,...,NMPPQ, I.NE.NUMP+1.
      x(nmpp1)=zero
      error(1:numgr)=zero
      RETURN
    END IF
    iflag=3333
    RETURN
  END IF
!
! HERE THE UNIMPROVED INITIAL APPROXIMATION WAS USER SUPPLIED OR
! COMPUTED BY THE LOEB INITIALIZATION.  SET THE ONES DIGIT OF
! IFLAG TO INDICATE THE INITIAL APPROXIMATION WAS NOT IMPROVED.
  iflag=-1
END IF
!
! IF IRET1 < 0 SET THE TENS DIGIT OF IFLAG TO INDICATE
! THAT THE TOLERANCES IN SLNPRO HAD TO BE ADJUSTED DURING THE
! COMPUTATION OF THE APPROXIMATION RETURNED.
  IF (iret1 < 0) iflag=iflag - 10

! NOW WE CHECK THE USER SUPPLIED CONSTRAINTS (IF ANY).
! CALL SPQERD TO RECOMPUTE THE VALUES OF P AND Q.  QMIN AND
! IQMIN ARE ALSO RECOMPUTED BY SPQERD.
  940 CALL spqerd (nump, numq, numgr, ftbwt, pwr, indf, wtble, x, v,  &
                   qmin, iqmin)
  icon=icnck(iopt, nump, numq, numgr, indlw, altbl, indup, uptbl, epsq, x, v,  &
             meq, mineq, alc, epres)
  IF (icon > 0) iflag=iflag-100

960 RETURN
END SUBROUTINE cdfcor



SUBROUTINE stcomp (numdm, nrwgr, nclgr, aswp1, aswp2, anep1, anep2, t)

!***BEGIN PROLOGUE  STCOMP
!***DATE WRITTEN   720120 (YYMMDD)
!***REVISION DATE  880226   (YYMMDD)
!***CATEGORY NO.  M2,Q
!***KEYWORDS  GRID CONSTRUCTION
!***AUTHOR  KAUFMAN, EDWIN H. JR.,
!             DEPARTMENT OF MATHEMATICS
!             CENTRAL MICHIGAN UNIVERSITY
!             MOUNT PLEASANT, MICHIGAN 48859
!           TAYLOR, GERALD D.,
!             DEPARTMENT OF MATHEMATICS
!             COLORADO STATE UNIVERSITY
!             FORT COLLINS, COLORADO 80523
!***PURPOSE  THIS SUBROUTINE, WHICH IS CALLED ONLY BY THE USER, CAN BE USED TO
!            COMPUTE THE COORDINATES OF THE GRID POINTS IF THEY ARE AN EQUALLY
!            SPACED (IN EACH DIRECTION) SUBSET OF AN INTERVAL OR RECTANGLE.
!***DESCRIPTION
!
!  NUMDM  (INPUT) THIS IS THE NUMBER OF DIMENSIONS (1 OR 2).
!
!  NRWGR  (INPUT) THIS NEED NOT BE ASSIGNED A VALUE IN THE
!         1 DIMENSIONAL CASE.  IN THE 2 DIMENSIONAL CASE IT
!         IS THE NUMBER OF ROWS OF THE GRID.
!
!  NCLGR  (INPUT) THIS IS THE NUMBER OF GRID POINTS IN THE 1
!         DIMENSIONAL CASE.  IN THE 2 DIMENSIONAL CASE IT
!         IS THE NUMBER OF COLUMNS OF THE GRID.
!
!  ASWP1  (INPUT) THIS IS THE LEFT END POINT IN THE 1 DIMENSIONAL
!         CASE.  IN THE 2 DIMENSIONAL CASE IT IS THE FIRST
!         COORDINATE OF THE LOWER LEFT CORNER POINT.
!
!  ASWP2  (INPUT) THIS IS THE RIGHT END POINT IN THE 1 DIMENSIONAL
!         CASE.  IN THE 2 DIMENSIONAL CASE IT IS THE SECOND
!         COORDINATE OF THE LOWER LEFT CORNER POINT.
!
!  ANEP1  (INPUT) THIS NEED NOT BE ASSIGNED A VALUE IN THE 1
!         DIMENSIONAL CASE.  IN THE 2 DIMENSIONAL CASE IT IS
!         THE FIRST COORDINATE OF THE UPPER RIGHT CORNER POINT.
!
!  ANEP2  (INPUT) THIS NEED NOT BE ASSIGNED A VALUE IN THE 1
!         DIMENSIONAL CASE.  IN THE 2 DIMENSIONAL CASE IT IS
!         THE SECOND COORDINATE OF THE UPPER RIGHT CORNER POINT.
!
!  T      (OUTPUT) IN THE 1 DIMENSIONAL CASE THE COORDINATE
!         OF THE ITH GRID POINT WILL BE PLACED IN T(I). IN
!         THE 2 DIMENSIONAL CASE THE COORDINATES OF THE ITH
!         GRID POINT (READING ROW BY ROW, TOP TO BOTTOM) WILL
!         BE PLACED IN (T(2*I-1),T(2*I)).
!***REFERENCES  (NONE)
!***ROUTINES CALLED  (NONE)
!***END PROLOGUE  STCOMP
!

INTEGER, INTENT(IN)     :: numdm
INTEGER, INTENT(IN)     :: nrwgr
INTEGER, INTENT(IN)     :: nclgr
REAL (dp), INTENT(IN)   :: aswp1
REAL (dp), INTENT(IN)   :: aswp2
REAL (dp), INTENT(IN)   :: anep1
REAL (dp), INTENT(IN)   :: anep2
REAL (dp), INTENT(OUT)  :: t(202)

!
REAL (dp)  :: hspac, vspac, xcor, ycor, zero
INTEGER    :: i, ind1, ind2, INDEX, j, ncgdb
!
! SET PRECISION DEPENDENT CONSTANTS FOR STCOMP.
!***FIRST EXECUTABLE STATEMENT  STCOMP
zero=0.0_dp
! END SETTING PRECISION DEPENDENT CONSTANTS FOR STCOMP.
!
IF (numdm /= 2) THEN
!
! THIS IS THE 1 DIMENSIONAL CASE.
  hspac=(aswp2-aswp1)/(nclgr-1)
  DO  i=1,nclgr
    t(i)=aswp1 + hspac*(i-1)
  END DO
  RETURN
END IF
!
! THIS IS THE 2 DIMENSIONAL CASE.
ncgdb=2*nclgr
IF (nclgr /= 1) THEN
  hspac=(anep1-aswp1)/(nclgr-1)
ELSE
  hspac=zero
END IF
IF (nrwgr /= 1) THEN
  vspac=(anep2-aswp2)/(nrwgr-1)
ELSE
  vspac=zero
END IF

DO  INDEX=1,nrwgr
  ycor=anep2 - vspac*(INDEX-1)
  ind1=2 + ncgdb*(INDEX-1)
  ind2=ind1 + ncgdb-2
! FILL IN THE Y COORDINATES IN ROW INDEX.
  DO  j=ind1,ind2,2
    t(j)=ycor
  END DO
END DO
DO  INDEX=1,nclgr
  xcor=aswp1 + hspac*(INDEX-1)
  ind1=2*INDEX-1
  ind2=ind1 + ncgdb*(nrwgr-1)
! FILL IN THE X COORDINATES IN COLUMN INDEX.
  DO  j=ind1,ind2,ncgdb
    t(j)=xcor
  END DO
END DO

RETURN
END SUBROUTINE stcomp



SUBROUTINE spqerd (nump, numq, numgr, ftbwt, pwr, indf, wtble, x, v, qmin, &
                   iqmin)
!***BEGIN PROLOGUE  SPQERD
!***REFER TO CDFCOR
!***ROUTINES CALLED  (NONE)
!***PURPOSE  THIS SUBROUTINE COMPUTES THE VALUES OF P, Q, AND
!            THE ERROR F-WT*P/Q, AS WELL AS THE UNIFORM ERROR NORM
!            AND THE MINIMUM OF THE DENOMINATOR, FOR THE
!            DIFFERENTIAL CORRECTION ALGORITHM.
!***END PROLOGUE  SPQERD
!

INTEGER, INTENT(IN)        :: nump
INTEGER, INTENT(IN)        :: numq
INTEGER, INTENT(IN)        :: numgr
REAL (dp), INTENT(IN)      :: ftbwt(101)
REAL (dp), INTENT(IN)      :: pwr(101,15)
INTEGER, INTENT(IN OUT)    :: indf(101)
REAL (dp), INTENT(IN)      :: wtble(101)
REAL (dp), INTENT(IN)      :: x(16)
REAL (dp), INTENT(IN OUT)  :: v(635,17)
REAL (dp), INTENT(OUT)     :: qmin
INTEGER, INTENT(OUT)       :: iqmin

!
! FOR EACH GRID POINT I THIS SUBROUTINE COMPUTES P,Q, AND
! THE ERROR F-WT*P/Q.  THE UNIFORM ERROR NORM
! MAX ABS(F-WT*P/Q) IS ALSO COMPUTED.  RECALL THAT IF WE ARE
! DOING WEIGHTED ERROR CURVE APPROXIMATION, F WILL ACTUALLY BE WT*F.
!
! SET MACHINE DEPENDENT PARAMETERS FOR SUBROUTINE SPQERD.
! SET SPCMN=B**(-T), WHERE B IS THE BASE AND T IS THE NUMBER
! OF BASE B DIGITS IN THE MANTISSA.  SPCMN IS THE MINIMUM
! RELATIVE SPACING ABS((X1-X2)/X2) BETWEEN TWO SUCCESSIVE
! FLOATING POINT NUMBERS, SO IT IS THE SPACING BETWEEN TWO
! SUCCESSIVE FLOATING POINT NUMBERS IN THE CLOSED INTERVAL
! (0.1,1.0).  WE ALSO HAVE SPCMN=10.0**(-T*(LOG10)(B))=
! 10.0**(-TNMAN), WHERE TNMAN IS THE BASE 10 EQUIVALENT OF
! THE LENGTH OF THE MANTISSA.
!
REAL (dp)  :: abad, aberr, eps2, four, one, spcmn, ten, two, zero
INTEGER    :: i, idb, idbm, j, n, nmpp1, nmppq, nn, np1

!***FIRST EXECUTABLE STATEMENT  SPQERD
spcmn=EPSILON(1.0_dp)

! SET PRECISION DEPENDENT CONSTANTS FOR SPQERD.
one=1.0D0
zero=0.0_dp
two=one+one
four=two+two
ten=four+four+two
! SET EPS2=100.0*SPCMN.  THUS EPS2=10.0**(-(TNMAN-2)).
! THE DENOMINATOR WILL NOT BE ALLOWED TO BE SMALLER THAN
! EPS2 AT ANY GRID POINT.  THIS WILL AVOID A DIVIDE FAULT
! AND ALSO REJECT SOME BAD APPROXIMATIONS.
!
eps2=ten*ten*spcmn
! END OF SETTING MACHINE AND PRECISION DEPENDENT PARAMETERS FOR SPQERD.
!
nmpp1=nump+1
nmppq=nump+numq
n=nmppq+1
np1=n+1
nn=2*numgr+1
i=0
idbm=-1
idb=0
v(nn,np1)=zero
abad=one
10 i=i+1
IF (i > numgr) THEN
!
! REPLACE V(NN,NP1) BY -V(NN,NP1)-2.0 IF THE DENOMINATOR AT
! SOME POINT WAS VERY SMALL IN ABSOLUTE VALUE OR NEGATIVE.
  v(nn,np1)=abad*v(nn,np1) + abad - ABS(abad)
  RETURN
END IF
!
idbm=idbm+2
idb=idb+2
!
! COMPUTE THE VALUE OF P AT GRID POINT I AND PUT IT IN
! V(2*I-1,N+1).
v(idbm,np1)=zero
DO  j=1,nump
  v(idbm,np1)=v(idbm,np1)+x(j)*pwr(i,j)
END DO
!
! COMPUTE THE VALUE OF Q AT GRID POINT I AND PUT IT IN V(2*I,N+1).
v(idb,np1)=zero
DO  j=nmpp1,nmppq
  v(idb,np1)=v(idb,np1) + x(j)*pwr(i,j)
END DO

! KEEP TRACK OF THE MINIMUM OF THE DENOMINATOR IN QMIN AND THE
! INDEX OF THE GRID POINT WHERE IT OCCURS IN IQMIN.
IF (i > 1) THEN
  IF (v(idb,np1) >= qmin) GO TO 80
END IF
qmin=v(idb,np1)
iqmin=i
80 IF (v(idb,np1) >= eps2) THEN
!
! NOW COMPUTE THE ERROR AT GRID POINT I AND PUT IT IN
! V(I,N).  KEEP TRACK OF THE ERROR NORM IN V(2*NUMGR+1,N+1).
! IF INDF(I)=0, THEN F IS TO BE IGNORED AT GRID POINT I,
! SO SET V(I,N)=0.0 AND LOOK AT THE NEXT GRID POINT.
  IF (indf(i) <= 0) THEN
    v(i,n)=zero
    GO TO 10
  END IF
  v(i,n)=ftbwt(i) - wtble(i)*v(idbm,np1)/v(idb,np1)
  aberr=ABS(v(i,n))
  IF (aberr <= v(nn,np1)) GO TO 10
  v(nn,np1)=aberr
  GO TO 10
END IF
!
! HERE V(2*I,N+1) IS VERY SMALL IN ABSOLUTE VALUE OR NEGATIVE.
! WE SET THE ERROR EQUAL TO ZERO AND REPLACE THE FINAL ERROR
! NORM BY ITS NEGATIVE, MINUS 2.0, AS A WARNING.  THIS
! SITUATION SOMETIMES OCCURS IF THE PREVIOUS APPROXIMATION
! WAS NEAR BEST, OR IF AN APPROXIMATION COMPUTED ON A COARSE
! GRID WAS USED TO INITIALIZE THE MAIN ALGORITHM ON A FINE GRID.
abad=-one
v(i,n)=zero

! TO INCREASE THE PROBABILITY THAT THE DENOMINATOR OF THE
! NEXT COMPUTED APPROXIMATION WILL BE >= EPS2 IF THE MAIN
! PROGRAM CONTINUES ON, WE RESET THE DENOMINATOR OF THE
! PRESENT APPROXIMATION AT THE BAD POINTS.
v(idb,np1)=ten*eps2
GO TO 10
END SUBROUTINE spqerd



SUBROUTINE slnpro (v, m, n, iyrct, x, indic)
!***BEGIN PROLOGUE  SLNPRO
!***REFER TO  CDFCOR
!***ROUTINES CALLED  SJELIM
!***PURPOSE  THIS SUBROUTINE SOLVES THE LINEAR PROGRAMMING PROBLEM
!              MAXIMIZE Z = -V(M+1,1)*X(1)-...-V(M+1,N)*X(N)
!              WHERE X(1),...,X(N) ARE FREE, SUBJECT TO
!              V(I,1)*X(1)+...+V(I,N)*X(N) <= V(I,N+1), FOR I=1,..,N.
!            (INFORMATION CONCERNING TOLERANCES AND BASIC VARIABLES
!            IS ALSO TRANSMITTED USING M, N, AND IYRCT.)
!***REFERENCES  AVDEYEVA, L. I. AND ZUKHOVITSKIY, S. I.,
!                 LINEAR AND CONVEX PROGRAMMING,
!                 SAUNDERS, PHILADELPHIA, 1966.
!***END PROLOGUE  SLNPRO
!

REAL (dp), INTENT(IN OUT)  :: v(635,17)
INTEGER, INTENT(IN OUT)    :: m
INTEGER, INTENT(OUT)       :: n
INTEGER, INTENT(IN OUT)    :: iyrct(634)
REAL (dp), INTENT(OUT)     :: x(16)
INTEGER, INTENT(OUT)       :: indic

!
REAL (dp)  :: y(634)
INTEGER    :: ixrct(634), iycct(16)
!
! GIVEN INTEGERS M AND N (WITH M>=N) AND A MATRIX V,
! THIS SUBROUTINE SOLVES THE LINEAR PROGRAMMING PROBLEM
!    MAXIMIZE Z=-V(M+1,1)X(1)-...-V(M+1,N)X(N)+V(M+1,N+1)
! SUBJECT TO THE CONSTRAINTS
!    V(I,1)X(1)+...+V(I,N)X(N)<=V(I,N+1), I=1,...,M
! USING ESSENTIALLY THE METHOD IN THE BOOK BY AVDEYEVA AND
! ZUKHOVITSKIY.  Y(I)=-V(I,1)X(1)-...-V(I,N)X(N)+V(I,N+1),
! I=1,...,M ARE SLACK VARIABLES.  THE METHOD HAS 4 PHASES.
!
! FIRST, XS ARE EXCHANGED FOR YS TO GET A PROBLEM
! INVOLVING ONLY NONNEGATIVE VARIABLES, THE YS BEING
! SELECTED IN THE ORDER DETERMINED BY IYRCT AND A PIVOTING
! STRATEGY.  AT THE BEGINNING OF THIS ROUTINE WE MUST HAVE
! IYRCT(1) NONPOSITIVE, OR IYRCT MUST CONTAIN SOME
! PERMUTATION OF THE INTEGERS 1,...,M (SEE BELOW).
! SECOND, THE SLACK CONSTANTS OF THE DUAL PROBLEM ARE MADE
! (SIGNIFICANTLY) NONNEGATIVE.
! THIRD, THE COST COEFFICIENTS OF THE DUAL PROBLEM ARE MADE
! (SIGNIFICANTLY) NONNEGATIVE.
! FINALLY, THE SOLUTION VECTOR IS COMPUTED.
!
! THE VARIABLE INDIC WILL BE GIVEN VALUE
! 0, IF A SOLUTION WAS FOUND NORMALLY
! 1, IF THERE WAS TROUBLE IN PHASE 1
! 2, IF THERE WAS TROUBLE IN PHASE 2 (EITHER ROUND OFF
!    ERROR, OR THE ORIGINAL PROBLEM WAS INCONSISTENT OR UNBOUNDED)
! 3, IF THERE WAS TROUBLE IN PHASE 3 (EITHER ROUND OFF
!    ERROR, OR THE ORIGINAL CONSTRAINTS WERE INCONSISTENT)
! 4, IF 300 MODIFIED JORDAN ELIMINATIONS WERE USED IN PHASES 2 AND 3
! -1, IF A SOLUTION WAS FOUND BUT IN ORDER TO OVERCOME
!   TROUBLE IN PHASE 2 OR 3 IT WAS NECESSARY TO TEMPORARILY
!   RELAX THE RESTRICTION ON DIVISION BY NUMBERS WITH SMALL
!   ABSOLUTE VALUE.  THUS THERE IS AN INCREASED CHANCE OF
!   SERIOUS ROUNDOFF ERROR IN THE RESULTS.
! -2, IF A SOLUTION WAS FOUND NORMALLY, EXCEPT THAT
!   THE PARAMETERS REA AND REA1 WERE INCREASED BY A SIGNAL
!   FROM THE CALLING PROGRAM (NAMELY, M=-M).  THE INCREASED
!   TOLERANCES MAY HAVE ALLOWED MORE ERROR THAN USUAL.
! -3, IF IN ORDER TO COMPLETE PHASE 1 IT WAS NECESSARY TO
!   TEMPORARILY RELAX THE RESTRICTION ON DIVISION BY NUMBERS
!   WITH SMALL ABSOLUTE VALUE.  THUS THERE IS AN INCREASED
!   CHANCE OF SERIOUS ROUNDOFF ERROR IN THE RESULTS.
! -4, IF A SOLUTION WAS FOUND NORMALLY, EXCEPT THAT REA AND REA1
!   WERE DECREASED BY A SIGNAL FROM THE CALLING PROGRAM (NAMELY
!   N=-N) IN ORDER TO TRY FOR MORE ACCURACY.  THIS INCREASES THE
!   CHANCES OF SERIOUS ROUNDOFF ERROR IN THE RESULTS.
! NOTE THAT INDIC=-3 WILL OVERWRITE (AND THUS CONCEAL) INDIC=-2
!   OR INDIC=-4, AND INDIC=-1 WILL OVERWRITE INDIC=-2, -3, OR -4
!
! SET MACHINE DEPENDENT PARAMETERS FOR SUBROUTINE SLNPRO.
! SET SPCMN=B**(-T), WHERE B IS THE BASE AND T IS THE NUMBER
! OF BASE B DIGITS IN THE MANTISSA.  SPCMN IS THE MINIMUM
! RELATIVE SPACING ABS((X1-X2)/X2) BETWEEN TWO SUCCESSIVE
! FLOATING POINT NUMBERS, SO IT IS THE SPACING BETWEEN TWO
! SUCCESSIVE FLOATING POINT NUMBERS IN THE CLOSED INTERVAL
! (0.1,1.0).  WE ALSO HAVE SPCMN=10.0**(-T*(LOG10)(B))=
! 10.0**(-TNMAN), WHERE TNMAN IS THE BASE 10 EQUIVALENT OF
! THE LENGTH OF THE MANTISSA.
!

REAL (dp)  :: absv, amax, amin, ampr2, amprv, bmpr2, bmprv, dist, dist1,  &
              four, one, rea, rea1, rea2, rea3, reakp, rowq, rtcol, spcmn, &
              temp, ten, two, zero
INTEGER    :: i, i1, i2, i10, i20, iback, id, ifail, ii, inamp, indst,  &
              irlax, irow, itemp, ixrj, iycj, iyri, iytmp, j, jj, k, &
              KEEP, keep1, kkk, kpmp1, kpmp2, ktjor, l, ll, lrknt, mp1,  &
              mxrkn, np1

!***FIRST EXECUTABLE STATEMENT  SLNPRO
spcmn=EPSILON(1.0_dp)

! SET PRECISION DEPENDENT CONSTANTS FOR SLNPRO.
one=1.0D0
zero=0.0_dp
two=one+one
four=two+two
ten=four+two+two

! SET REA (ROUND OFF ERROR ADJUSTMENT) =
! MAX(10.0**(-8),100.0*SPCMN).  THUS REA=10.0**(-EXREA),
! WHERE EXREA=MIN(8,TNMAN-2).
! DIVISION BY NUMBERS <= REA IN ABSOLUTE VALUE WILL NOT BE PERMITTED.
rea=ten*ten*spcmn
IF (rea-ten**(-8) < 0.0) rea=ten**(-8)

! SET REA1=10.0*SPCMN.  THUS REA1=10.0**(-(TNMAN-1)).
! NUMBERS IN ROW M+1 OR COLUMN N+1 WHICH ARE <= REA1 IN
! ABSOLUTE VALUE WILL BE TREATED AS ZEROES.  SLNPRO ASSUMES
! THAT 0.0<REA1<=REA.
rea1=ten*spcmn
! END OF INITIAL SETTING OF MACHINE DEPENDENT PARAMETERS FOR
! SLNPRO.  THESE PARAMETERS MAY BE ADJUSTED BY A COMMAND FROM CDFCOR.
!
indic=0
! M BEING NEGATIVE IS A SIGNAL TO INCREASE REA AND REA1,
! THUS TREATING MORE NUMBERS WITH SMALL ABSOLUTE VALUES AS
! ZEROES.  THIS MAY GIVE THIS ROUTINE A BETTER CHANCE TO
! SUCCEED, BUT MAY ALSO CAUSE LARGER ERRORS.
IF (m < 0) THEN

! RESET M.
  m=-m
  rea=SQRT(rea)
  rea1=SQRT(rea1)
  indic=-2
END IF

! N BEING NEGATIVE IS A SIGNAL TO DECREASE REA AND REA1 TO TRY
! FOR MORE ACCURACY.  AMONG OTHER THINGS, THIS MAKES IT MORE
! LIKELY THAT THE PREVIOUS VERTEX WILL BE RETAINED IN PHASE 1
! BELOW, BUT IT ALSO COULD INCREASE ROUND OFF ERROR.
IF (n < 0) THEN

! RESET N.
  n=-n
  rea=rea1
  rea1=rea1/(ten*ten)
  indic=-4
END IF

! PRESERVE REA IN CASE IT MUST BE TEMPORARILY RELAXED.
! IRLAX=0 INDICATES REA IS NOT RELAXED AT THIS STAGE.
reakp=rea
irlax=0
! IN COLUMN N+1, NUMBERS <= REA2 IN ABSOLUTE VALUE WILL BE TREATED AS ZEROES.
rea2=rea1
np1=n+1
mp1=m+1
ktjor=0
iback=0
! SET V(MP1,NP1)=0.0 SO THE DESCRIPTIONS IN AND FOLLOWING THE PROLOGUE
! WILL AGREE.
v(mp1,np1)=zero

! THE ONLY REASON FOR THE FOLLOWING THREE STATEMENTS IS TO
! AVOID THE ERROR MESSAGE ON SOME MACHINES THAT THESE
! VARIABLES HAVE NOT BEEN ASSIGNED A VALUE.  THEY WILL BE
! REASSIGNED A VALUE BEFORE THE PROGRAM REACHES A SPOT WHERE
! THEY WILL ACTUALLY BE USED.
dist=one
amprv=one
ampr2=one

! SET IXRCT.  IXRCT(I)=0 MEANS SOME Y IS IN ROW I, WHILE
! IXRCT(I)=K.NE.0 MEANS X(K) IS IN ROW I.
ixrct(1:m)=0
!
! EXCHANGE THE XS AT THE TOP OF THE TABLE FOR YS.
! IF IYRCT(1) IS NONPOSITIVE, WE SET IYRCT AND CHOOSE THE
! LARGEST POSSIBLE RESOLVENTS FOR THE EXCHANGES.  IF
! IYRCT(1) IS POSITIVE, IYRCT WILL HAVE BEEN PREVIOUSLY SET
! AND WE TRY TO EXCHANGE IN ROWS IYRCT(1),...,IYRCT(N),
! STILL EMPLOYING A PIVOTING STRATEGY, BUT IF WE CANNOT, WE
! EXCHANGE IN ROWS IYRCT(N+1),...,IYRCT(M).
IF (iyrct(1) <= 0) THEN
  i10=1
  i20=m

! IF WE HAVE NO INFORMATION FROM A PREVIOUS VERTEX, WE GIVE
! UP A LITTLE ACCURACY IN COLUMN N+1 TO HAVE A BETTER CHANCE OF SUCCESS.
  rea2=rea

! THIS ROUTINE HAS A BACKTRACKING OPTION WHICH SOMETIMES
! INCREASES ACCURACY BUT SOMETIMES LEADS TO FAILURE DUE TO
! CYCLING.  IT IS SUGGESTED THAT THIS OPTION BE EMPLOYED IF
! INFORMATION ABOUT A STARTING VERTEX IS AVAILABLE, AND
! OTHERWISE BE DISABLED BY SETTING IBACK=1.
  iback=1
  DO  i=1,m
    iyrct(i)=i
  END DO
ELSE
  i10=1
  i20=n
END IF
j=0

! SET THE LOWER BOUND ON THE ABSOLUTE VALUE OF A RESOLVENT IN
! PHASE 1.  ALSO SET IFAIL=0 TO INDICATE THE RESOLVENT SEARCH
! IN THIS COLUMN HAS NOT FAILED.
rea3=rea
ifail=0
120 j=j+1
IF (j <= n) THEN

! SET I1, I2 ACCORDING TO THE STRATEGY WE ARE USING.
  130 i1=i10
  i2=i20
  amax=zero

! SEARCH FOR A RESOLVENT IN ROWS IYRCT(I1),...,IYRCT(I2).
  140 DO  i=i1,i2
    iytmp=iyrct(i)
    IF (ixrct(iytmp) /= 0) CYCLE
    absv=ABS(v(iytmp,j))
    IF (absv <= amax) CYCLE
    iyri=iytmp
    amax=absv
  END DO

! CHECK TO SEE IF THE PROSPECTIVE RESOLVENT IS LARGE ENOUGH IN ABSOLUTE VALUE.
  IF (amax > rea3) THEN
! EXCHANGE X(J) FOR Y(IYRI).
    CALL sjelim (mp1, 1, np1, iyri, j, v)
    ixrct(iyri)=j
    iycct(j)=iyri

! IYCCT(J)=IYRI MEANS Y(IYRI) IS IN COLUMN J.
! RESET REA3 AND IFAIL SINCE WE SUCCESSFULLY FOUND A RESOLVENT IN
! THIS COLUMN, AND THE RESOLVENT SEARCH IN THE NEXT COLUMN HAS NOT FAILED.
    rea3=rea
    ifail=0
    GO TO 120
  END IF

! WE HAVE NOT FOUND A SUITABLE RESOLVENT IN ROWS IYRCT(I1),
! ...IYRCT(I2).  IF I2 < M WE SEARCH THE REST OF COLUMN J.
  IF (i2 < m) THEN
    i1=i2+1
    i2=m
    GO TO 140
  END IF

! HERE WE FAILED TO FIND A RESOLVENT IN COLUMN J WITH ABSOLUTE
! VALUE > REA3.  IF IFAIL=0 WE SET INDIC=-3 AND TRY AGAIN
! WITH REA3 REDUCED.  IF THIS HAS ALREADY BEEN TRIED WE SET
! INDIC=1 AND RETURN.
  IF (ifail <= 0) THEN
    ifail=1
    indic=-3
    rea3=rea1
    GO TO 130
  END IF

  indic=1
  RETURN
END IF
!
! REARRANGE THE ROWS OF V SO THAT X(1),...,X(N) COME FIRST
! IN THAT ORDER.  REDEFINE IYRCT SO THAT AFTER THE
! REARRANGEMENT IS DONE, IYRCT(I)=K WILL MEAN Y(K) IS IN
! ROW I (FOR I GREATER THAN N).
DO  i=1,m
  iyrct(i)=i
END DO
irow=0
260 irow=irow+1
IF (irow <= m) THEN
  IF (ixrct(irow) == 0) GO TO 260
  IF (ixrct(irow) == irow) GO TO 260

! NOW X(L) IS IN ROW IROW, BUT WE WANT IT IN ROW L.
  290 l=ixrct(irow)
  ll=ixrct(l)
  IF (ll /= 0) THEN

! X(L) IS IN ROW IROW, WHILE X(LL) IS IN ROW L.
    ixrct(irow)=ll
    ixrct(l)=l
  ELSE

! X(L) IS IN ROW IROW, WHILE Y(IYRCT(L)) IS IN ROW L.
    ixrct(irow)=0
    iyrct(irow)=iyrct(l)
    ixrct(l)=l
  END IF

! NOW EXCHANGE THE CONTENTS OF ROWS IROW AND L.
  DO  j=1,np1
    temp=v(irow,j)
    v(irow,j)=v(l,j)
    v(l,j)=temp
  END DO
  IF (ixrct(irow) == 0) GO TO 260
  IF (ixrct(irow) == irow) GO TO 260
  GO TO 290
END IF

! NOW IXRCT IS NO LONGER NEEDED, SO STORE THE PRESENT IYCCT IN IT.
ixrct(1:n)=iycct(1:n)
! END OF PHASE 1.
!
! THE FIRST N ROWS OF V GIVE THE XS IN TERMS OF CERTAIN
! YS.  THESE ROWS WILL NOT BE CHANGED BY LATER OPERATIONS.
!
! WE NOW ATTACK THE MAXIMIZATION PROBLEM BY LOOKING AT THE
! DUAL PROBLEM OF MINIMIZING A FORM GIVEN BY THE
! COEFFICIENTS IN V(N+1,N+1) THROUGH V(M,N+1) WITH SLACK
! TERMS IN THE BOTTOM ROW OF V.
! SEARCH ROW M+1 FOR A SIGNIFICANTLY NEGATIVE ELEMENT.  IF
! THERE ARE NONE, PROCEED TO THE ACTUAL MINIMIZATION
! PROBLEM.  STICK WITH COLUMN JJ UNTIL V(M+1,JJ)>=-REA1.
jj=0
370 jj=jj+1
IF (jj > n) GO TO 760
IF (v(mp1,jj) + rea1 >= 0.0) GO TO 370
!
! WE HAVE V(M+1,JJ) SIGNIFICANTLY NEGATIVE.  SEARCH COLUMN JJ FOR A
! POSITIVE ELEMENT, TREATING A VERY SMALL V(I,J) AS A ZERO.
! IF THERE ARE NO POSITIVE ELEMENTS THE DUAL CONSTRAINTS WERE INCONSISTENT,
! SO THE ORIGINAL PROBLEM WAS INCONSISTENT OR UNBOUNDED.
390 i=n
inamp=0
400 i=i+1
IF (i <= m) THEN
  IF (v(i,jj) <= rea) GO TO 400
!
! NOW V(I,JJ) > REA.  WE SEARCH ROW I FOR INDICES K SUCH
! THAT V(M+1,K) >= 0.0 .OR. K<JJ, AND V(I,K) < -REA, AND
! FIND THE MAXIMUM RATIO (I.E. THE RATIO WITH SMALLEST
! ABSOLUTE VALUE, IF V(M+1,K) >= 0.0) V(M+1,K)/V(I,K).  IF
! THERE IS NO SUCH K WE LOOK AT POSITIVE V(I,K) BELOW.
  indst=0
  DO  j=1,n
    IF (v(mp1,j) < 0.0) THEN
      IF (j >= jj) CYCLE
    END IF
    IF (v(i,j) + rea < 0.0) THEN
      dist1=v(mp1,j)/v(i,j)
      IF (indst > 0) THEN
        IF (dist1 <= dist) CYCLE
      END IF
      dist=dist1
      indst=1
      k=j
    END IF
  END DO
  IF (indst > 0) THEN
!
! WE NOW COMPUTE V(I,JJ)*DIST AND GO ON TO LOOK AT OTHER ROWS
! TO MINIMIZE THIS QUANTITY (I.E. TO MAXIMIZE ITS ABSOLUTE VALUE,
! IF V(M+1,K)>=0.0).  THIS IS THE NEGATIVE OF THE CHANGE IN V(M+1,JJ).
    bmprv=v(i,jj)*dist
    IF (inamp > 0) THEN
      IF (bmprv >= amprv) GO TO 400
    END IF
    amprv=bmprv
    inamp=1
    kpmp1=i
    kpmp2=k
! (KPMP1,KPMP2) GIVES THE POSITION OF THE BEST PROSPECTIVE
! RESOLVENT FOUND SO FAR.
    GO TO 400
  END IF
!
! IF THERE WAS NO INDEX K SUCH THAT V(M+1,K) >= 0.0 .OR. K < JJ,
! AND V(I,K)<-REA, WE LOOK FOR THE SMALLEST (I.E. LARGEST IN ABSOLUTE VALUE)
! RATIO V(M+1,K)/V(I,K) FOR V(I,K) > REA AND V(M+1,K) < 0.0, AND PERFORM AN
! ELIMINATION WITH RESOLVENT V(I,K).  THERE IS AT LEAST ONE SUCH K, NAMELY JJ.
! THIS WILL FINISH PHASE 2 UNLESS BACKTRACKING IS NECESSARY.
  dist=one
  DO  j=1,n
    IF (v(mp1,j) >= 0.0) CYCLE
    IF (v(i,j) <= rea) CYCLE
    dist1=v(mp1,j)/v(i,j)
    IF (dist1 >= dist) CYCLE
    dist=dist1
    k=j
  END DO

ELSE
  IF (inamp <= 0) THEN

! AT THIS POINT INAMP IS POSITIVE IFF THERE WAS AT LEAST ONE
! ELEMENT > REA IN COLUMN JJ.  IF THERE WERE NONE, WE
! TEMPORARILY RELAX REA AND TRY AGAIN.
    IF (irlax <= 0) THEN
      irlax=1
      indic=-1
      rea=rea1
      GO TO 390
    END IF

    indic=2
    RETURN
  END IF

! CHECK TO SEE IF V(MP1,KPMP2) IS VERY SMALL IN ABSOLUTE
! VALUE OR NEGATIVE.  THIS INDICATES DEGENERACY.
  IF (v(mp1,kpmp2) <= rea) GO TO 630
! DO AN ELIMINATION WITH RESOLVENT V(KPMP1,KPMP2).
  620 i=kpmp1
  k=kpmp2
  GO TO 740
!
! WE ARE NOW STUCK IN DEGENERATE COLUMN KPMP2.  WE SEARCH
! EACH DEGENERATE COLUMN IN WHICH WE ARE STUCK FOR A
! RESOLVENT WHICH WILL KEEP US FROM GETTING STUCK IN THIS
! COLUMN NEXT TIME, AND TO REDUCE THE ROUND-OFF ERROR WE
! TAKE THE SMALLEST OF THESE (I.E. LARGEST IN ABSOLUTE
! VALUE) AS OUR ACTUAL RESOLVENT.
  630 amin=one
  DO  j=1,n
! COLUMN J MAY BE DEGENERATE IF 0.0 <= V(M+1,J) <= REA,
! OR V(M+1,J) < 0.0 .AND. J < JJ.
    IF (v(mp1,j) < 0.0) THEN
      IF (j >= jj) CYCLE
      GO TO 640
    END IF
    IF (v(mp1,j) > rea) CYCLE

! WE WILL BE STUCK IN COLUMN J IFF THERE IS AN INDEX ID FOR
! WHICH V(ID,JJ)>REA AND V(ID,J)<-REA.  IF THIS IS THE
! CASE, CHOOSING SUCH AN ID SO THAT V(ID,JJ)/V(ID,J) IS
! MINIMIZED (I.E. MAXIMIZED IN ABSOLUTE VALUE) AND TAKING
! V(ID,J) AS THE RESOLVENT WILL INSURE THAT WE DONT GET
! STUCK IN COLUMN J NEXT TIME.
    640 dist=one
    DO  i=np1,m
      IF (v(i,jj) <= rea) CYCLE
      IF (v(i,j) >= -rea) CYCLE
      dist1=v(i,jj)/v(i,j)
      IF (dist1 >= dist) CYCLE
      dist=dist1
      id=i
    END DO
    IF (dist >= 0.5_dp) CYCLE

! WE HAVE NOW DETERMINED THAT WE ARE STUCK IN COLUMN J.
! IF V(ID,J) < AMIN THEN V(ID,J) IS THE BEST RESOLVENT FOUND SO FAR.
    IF (v(id,j) >= amin) CYCLE
    amin=v(id,j)
    kpmp1=id
    kpmp2=j
  END DO

! THE BEST RESOLVENT IS V(KPMP1,KPMP2), SO WE DO AN ELIMINATION.
  GO TO 620
END IF

740 ktjor=ktjor+1
IF (ktjor > 300) GO TO 1270
CALL sjelim (mp1, np1, np1, i, k, v)
itemp=iyrct(i)
iyrct(i)=iycct(k)
iycct(k)=itemp

! RESET REA AND IRLAX.
rea=reakp
irlax=0

! IF NOW V(M+1,JJ) HAS BEEN MADE NOT SIGNIFICANTLY NEGATIVE,
! WE GO TO THE NEXT COLUMN.  OTHERWISE WE TRY AGAIN IN COLUMN JJ.
IF (v(mp1,jj)+rea1 < 0.0) THEN
  GO TO 390
ELSE
  GO TO 370
END IF

! IN THE UNLIKELY EVENT THAT SOME V(M+1,J) IS STILL VERY
! SIGNIFICANTLY NEGATIVE WE BACKTRACK TO COLUMN J.  THIS
! COULD NOT HAPPEN IF THERE WERE NO ROUNDOFF ERROR AND WE
! COULD ALLOW DIVISION BY NUMBERS WITH VERY SMALL ABSOLUTE
! VALUE.  OMIT BACKTRACKING IF IBACK=1.
760 IF (iback <= 0) THEN
  DO  j=1,n
    IF (v(mp1,j) > -rea) CYCLE
    jj=j
    GO TO 390
  END DO
END IF
! END OF PHASE 2.
!
800 i=n
kkk=0
!
! SEARCH FOR A SIGNIFICANTLY NEGATIVE ELEMENT BETWEEN
! V(N+1,N+1) AND V(N+1,M).  IF THERE ARE NONE WE HAVE THE
! MINIMAL POINT OF THE DUAL PROBLEM (AND THUS THE MAXIMAL
! POINT OF THE DIRECT PROBLEM) ALREADY.
810 i=i+1
IF (i <= m) THEN
  IF (v(i,np1) >= -rea2) GO TO 810
!
! SEARCH FOR A NEGATIVE ELEMENT IN ROW I, TREATING A NUMBER
! WHICH IS VERY SMALL IN ABSOLUTE VALUE AS A ZERO.  IF THERE
! ARE NO NEGATIVE ELEMENTS THE DUAL PROBLEM WAS UNBOUNDED
! BELOW, SO THE ORIGINAL CONSTRAINTS WERE INCONSISTENT.
! FIND THE INDEX K OF THE LARGEST (I.E. SMALLEST IN ABSOLUTE
! VALUE, IF V(M+1,K) >= 0.0) RATIO V(M+1,K)/V(I,K) WITH V(I,K) < -REA.
  830 indst=0
  DO  j=1,n
    IF (v(i,j) + rea >= 0.0) CYCLE
    dist1=v(mp1,j)/v(i,j)
    IF (indst > 0) THEN
      IF (dist1 <= dist) CYCLE
    END IF
    k=j
    indst=1
    dist=dist1
  END DO
  IF (indst <= 0) THEN

! RELAX REA AND LOOK FOR NEGATIVE ELEMENTS WITH SMALLER ABSOLUTE VALUE.
    IF (irlax <= 0) THEN
      irlax=1
      indic=-1
      rea=rea1
      GO TO 830
    END IF

    indic=3
    RETURN
  END IF
!
! COMPUTE THE IMPROVEMENT DIST*V(I,N+1) IN THE VALUE OF THE
! FORM USING V(I,K) AS THE RESOLVENT.  SET KKK=1 TO INDICATE
! A SIGNIFICANTLY NEGATIVE V(I,N+1) WAS FOUND, AND LOOK AT
! THE OTHER ROWS TO FIND THE RESOLVENT GIVING THE LARGEST IMPROVEMENT.
  bmpr2=dist*v(i,np1)

! RESET IRLAX SO THAT THE NEXT ROW WHICH NEEDS RELAXING DOES NOT TERMINATE
! THE ROUTINE.  REA WILL REMAIN RELAXED UNTIL AFTER THE NEXT ELIMINATION.
  irlax=0
  IF (kkk > 0) THEN
    IF (bmpr2 <= ampr2) GO TO 810
  END IF
  kkk=1
  KEEP=i
  keep1=k
  ampr2=bmpr2
  GO TO 810
END IF

! KKK=0 HERE IFF NONE OF THE COST COEFFICIENTS ARE SIGNIFICANTLY NEGATIVE.
IF (kkk == 0) THEN

! CHECK TO SEE IF ANY OF THE NUMBERS IN THE BOTTOM ROW HAVE
! BECOME VERY SIGNIFICANTLY NEGATIVE.  IF SO, WE MUST
! BACKTRACK TO PHASE 2 (SEE COMMENT ABOVE STATEMENT 1035).
! OMIT BACKTRACKING IF IBACK=1.
  IF (iback > 0) GO TO 1280
  DO  j=1,n
    IF (v(mp1,j) + rea > 0.0) CYCLE
    jj=j
    GO TO 390
  END DO
  GO TO 1280
END IF

! CHECK TO SEE IF V(MP1,KEEP1) IS VERY SMALL IN ABSOLUTE
! VALUE OR NEGATIVE.  THIS INDICATES DEGENERACY.
IF (v(mp1,keep1)-rea > 0.0) THEN
  GO TO 1000
ELSE
  GO TO 1010
END IF
! DO AN ELIMINATION WITH RESOLVENT V(KEEP,KEEP1).
1000 i=KEEP
k=keep1
GO TO 1250
!
! WE ARE NOW STUCK IN DEGENERATE COLUMN KEEP1.  WE SEARCH
! EACH DEGENERATE COLUMN IN WHICH WE ARE STUCK FOR A
! RESOLVENT WHICH WILL KEEP US FROM GETTING STUCK IN THIS
! COLUMN NEXT TIME.  IF WE ARE NOT USING THE OPTION
! DESCRIBED IN THE COMMENTS PRECEDING STATEMENT 1055, WE
! TAKE THE SMALLEST OF THESE (I.E. THE LARGEST IN ABSOLUTE
! VALUE) AS OUR ACTUAL RESOLVENT IN ORDER TO REDUCE THE
! GROWTH OF ROUND-OFF ERROR.
1010 amin=one
mxrkn=np1
DO  j=1,n
! COLUMN J MAY BE DEGENERATE IF V(M+1,J) <= REA.
  IF (v(mp1,j) > rea) CYCLE

! WE WILL BE STUCK IN COLUMN J IFF THERE IS AN INDEX ID FOR
! WHICH V(ID,N+1) < -REA2 AND V(ID,J) < -REA.  IF THIS
! IS THE CASE, CHOOSING SUCH AN ID SO THAT V(ID,N+1)/V(ID,J)
! IS MAXIMIZED AND TAKING V(ID,J) AS THE RESOLVENT WILL
! INSURE THAT WE DONT GET STUCK IN COLUMN J NEXT TIME.
  dist=-one
  DO  i=np1,m
    IF (v(i,np1) + rea2 >= 0.0) CYCLE
    IF (v(i,j) + rea >= 0.0) CYCLE
    dist1=v(i,np1)/v(i,j)
    IF (dist1 <= dist) CYCLE
    dist=dist1
    id=i
  END DO
  IF (dist <= -0.5_dp) CYCLE

! WE HAVE NOW DETERMINED THAT WE ARE STUCK IN COLUMN J.
! THE FOLLOWING STATEMENTS ATTEMPT TO BREAK DEGENERACY
! FASTER BY LOOKING ONE ITERATION INTO THE FUTURE, THAT IS,
! BY CHOOSING FROM THE PROSPECTIVE RESOLVENTS FOUND ABOVE
! THAT ONE WHICH MINIMIZES THE MINIMUM NUMBER OF STICKING
! PLACES IN ANY ROW AT THE NEXT STAGE.
! BECAUSE OF COMPUTER TIME AND THE POSSIBLE LOSS OF ACCURACY
! DUE TO LESSENED PIVOTING (EVEN THOUGH TIES ARE ALWAYS
! BROKEN IN FAVOR OF THE RESOLVENT WITH GREATEST ABSOLUTE
! VALUE), IT IS SUGGESTED THAT THIS OPTION BE OMITTED IF
! INFORMATION WAS AVAILABLE FROM A PREVIOUS VERTEX.  THIS
! WILL BE THE CASE IFF THE BACKTRACKING OPTION WAS USED,
! THAT IS, IFF IBACK=0.
  IF (iback > 0) THEN

! COMPUTE WHAT THE NEW BOTTOM ROW WOULD BE (EXCEPT FOR
! POSITION J) IF V(ID,J) WERE USED AS THE RESOLVENT, AND
! PUT THE RESULTS INTO Y.
    rowq=v(mp1,j)/v(id,j)
    DO  l=1,n
      IF (l == j) CYCLE
      y(l)=v(mp1,l) - v(id,l)*rowq
    END DO
    lrknt=-1

! WE LOOK FOR A ROW WHICH WILL HAVE A SIGNIFICANTLY NEGATIVE
! LAST ELEMENT BUT A MINIMUM NUMBER OF PLACES WHERE WE WILL
! BE STUCK IN DEGENERATE COLUMNS.  LRKNT=-1 MEANS WE HAVE
! NOT YET FOUND A ROW WHICH WILL HAVE A SIGNIFICANTLY
! NEGATIVE LAST ELEMENT.
    loop1200:  &
    DO  ii=np1,m
      IF (ii == id) CYCLE
      rowq=v(ii,j)/v(id,j)
      rtcol=v(ii,np1)-v(id,np1)*rowq
      IF (rtcol + rea2 >= 0.0) CYCLE

! IF WE HAVE ALREADY LOCATED A RESOLVENT WHICH WILL FINISH
! THE ROUTINE, BUT THE PRESENT PROSPECTIVE RESOLVENT WOULD
! GIVE A ROW WITH A SIGNIFICANTLY NEGATIVE LAST ELEMENT, WE
! LOOK AT THE NEXT PROSPECTIVE RESOLVENT FOR PIVOTING PURPOSES.
      IF (mxrkn == -1) CYCLE loop1200
      lrknt=0
! NOW COUNT THE NUMBER (LRKNT) OF STICKING PLACES IN ROW II
! AT THE NEXT ITERATION.
      DO  jj=1,n
        IF (jj == j) CYCLE
        IF (y(jj) > rea) CYCLE
        IF (v(ii,jj) - v(id,jj)*rowq + rea >= 0.0) CYCLE
        lrknt=lrknt+1
        IF (lrknt > mxrkn) CYCLE loop1200
      END DO

      IF (lrknt < mxrkn) THEN
        GO TO 1190
      ELSE IF (lrknt > mxrkn) THEN
        CYCLE
      END IF
      IF (v(id,j) >= amin) CYCLE
      1190 mxrkn=lrknt
      amin=v(id,j)
      KEEP=id
      keep1=j
    END DO loop1200

! LRKNT=-1 HERE WOULD MEAN THIS RESOLVENT WOULD FINISH THE ROUTINE.
! IF LRKNT>=0 THEN MXRKN>=0 ALSO, SO WE WILL
! NOT HAVE EARLIER FOUND A RESOLVENT WHICH WILL FINISH THE ROUTINE.
    IF (lrknt /= -1) CYCLE
    IF (mxrkn /= -1) GO TO 1230
  END IF
  IF (v(id,j) >= amin) CYCLE
  1230 mxrkn=-1
  amin=v(id,j)
  KEEP=id
  keep1=j
END DO

! THE BEST RESOLVENT IS V(KEEP,KEEP1), SO WE DO AN ELIMINATION.
GO TO 1000

1250 ktjor=ktjor+1
IF (ktjor > 300) GO TO 1270

CALL sjelim (mp1, np1, np1, i, k, v)
itemp=iyrct(i)
iyrct(i)=iycct(k)
iycct(k)=itemp

! RESET REA AND IRLAX.
rea=reakp
irlax=0
GO TO 800
1270 indic=4
RETURN

! END OF PHASE 3.  WE NOW HAVE THE VERTEX WE ARE SEEKING.
!
! READ OFF THE Y VALUES FOR THIS VERTEX.
1280 DO  j=1,n
  iycj=iycct(j)
  y(iycj)=zero
END DO
DO  i=np1,m
  iyri=iyrct(i)
  y(iyri)=v(i,np1)
END DO

! COMPUTE THE XS FROM THE YS.  RECALL THAT IXRCT CONTAINS THE FORMER IYCCT.
DO  i=1,n
  x(i)=v(i,np1)
  DO  j=1,n
    ixrj=ixrct(j)
    x(i)=x(i) - v(i,j)*y(ixrj)
  END DO
END DO
!
! NOW PUT THE VALUES IN IYCCT INTO THE FIRST N POSITIONS OF
! IYRCT IN DECREASING ORDER SO THAT WHEN SLNPRO IS CALLED
! AGAIN THE YS AT THE PRESENT MINIMIZING VERTEX WILL BE
! SCANNED FIRST, BEGINNING WITH THOSE CORRESPONDING TO
! -1.0<=Q(J)<=1.0.  THUS THESE WILL PROBABLY APPEAR IN
! THE INITIAL VERTEX AFTER EXCHANGE OF XS AND YS.
! TO ACCOMPLISH THIS, MAKE IXRCT(I)=-1 IF IYCCT(J)=I FOR
! SOME J, THEN SCAN IXRCT BACKWARDS.
DO  j=1,n
  iycj=iycct(j)
  ixrct(iycj)=-1
END DO
k=1
i=mp1
1340 i=i-1
IF (i > 0) THEN
  IF (ixrct(i) /= -1) GO TO 1340
  iyrct(k)=i
  k=k+1
  GO TO 1340
END IF

! NOW FILL IN THE REST OF IYRCT BY SCANNING IXRCT AGAIN.
i=mp1
1380 i=i-1
IF (i > 0) THEN
  IF (ixrct(i) < 0) GO TO 1380
  iyrct(k)=i
  k=k+1
  GO TO 1380
END IF

RETURN
END SUBROUTINE slnpro



SUBROUTINE sjelim (l, ll, k, ir, is, v)
!***BEGIN PROLOGUE  SJELIM
!***REFER TO  SLNPRO
!***ROUTINES CALLED  (NONE)
!***PURPOSE  THIS SUBROUTINE PERFORMS A MODIFIED JORDAN
!            ELIMINATION ON THE L-LL+1 BY K MATRIX
!            CONSISTING OF ROWS LL THROUGH L OF V AND
!            COLUMNS 1 THROUGH K OF V.  THE RESOLVENT IS V(IR,IS).
!***END PROLOGUE  SJELIM
!

INTEGER, INTENT(IN)        :: l
INTEGER, INTENT(IN)        :: ll
INTEGER, INTENT(IN)        :: k
INTEGER, INTENT(IN)        :: ir
INTEGER, INTENT(IN)        :: is
REAL (dp), INTENT(IN OUT)  :: v(635,17)
!

REAL (dp)  :: fact, one, resol
INTEGER    :: i, j
!
! SET PRECISION DEPENDENT CONSTANTS FOR SJELIM.
!***FIRST EXECUTABLE STATEMENT  SJELIM
one=1.0D0
! EMD OF SETTING PRECISION DEPENDENT CONSTANTS FOR SJELIM.
!
! DIVIDE THE ENTRIES IN THE RESOLVENT ROW (EXCEPT FOR THE
! RESOLVENT) BY THE RESOLVENT.
resol=v(ir,is)
DO  j=1,k
  IF (j /= is) v(ir,j)=v(ir,j)/resol
END DO

! SWEEP OUT IN ALL BUT ROW IR AND COLUMN IS.
DO  i=ll,l
  IF (i == ir) CYCLE
  fact=-v(i,is)
  DO  j=1,k
    IF (j /= is) v(i,j)=v(i,j) + v(ir,j)*fact
  END DO
END DO

! DIVIDE THE ENTRIES IN THE RESOLVENT COLUMN (EXCEPT FOR THE
! RESOLVENT) BY THE NEGATIVE OF THE RESOLVENT.
DO  i=ll,l
  IF (i /= ir) v(i,is)=-v(i,is)/resol
END DO

! REPLACE THE RESOLVENT BY ITS RECIPROCAL.
v(ir,is)=one/resol

RETURN
END SUBROUTINE sjelim




SUBROUTINE ssetv (nump, numq, numgr, ftbwt, pwr, iopt, epsq, indlw, altbl, &
                  indup, uptbl, indf, wtble, meq, mineq, alc, mode, v, m)
!***BEGIN PROLOGUE  SSETV
!***REFER TO  CDFCOR
!***ROUTINES CALLED  (NONE)
!***PURPOSE  THIS SUBROUTINE SETS UP THE LINEAR PROGRAMMING PROBLEM
!            FOR THE INITIAL (IF MODE=0) OR MAIN (IF MODE=1)
!            ITERATIONS OF THE DIFFERENTIAL CORRECTION ALGORITHM.
!***END PROLOGUE  SSETV2
!

INTEGER, INTENT(IN)        :: nump
INTEGER, INTENT(IN)        :: numq
INTEGER, INTENT(IN)        :: numgr
REAL (dp), INTENT(IN)      :: ftbwt(101)
REAL (dp), INTENT(IN)      :: pwr(101,15)
INTEGER, INTENT(IN)        :: iopt
REAL (dp), INTENT(IN)      :: epsq
INTEGER, INTENT(IN)        :: indlw(101)
REAL (dp), INTENT(IN)      :: altbl(101)
INTEGER, INTENT(IN)        :: indup(101)
REAL (dp), INTENT(IN)      :: uptbl(101)
INTEGER, INTENT(IN)        :: indf(101)
REAL (dp), INTENT(IN)      :: wtble(101)
INTEGER, INTENT(IN)        :: meq
INTEGER, INTENT(IN)        :: mineq
REAL (dp), INTENT(IN)      :: alc(101,16)
INTEGER, INTENT(IN)        :: mode
REAL (dp), INTENT(IN OUT)  :: v(635,17)
INTEGER, INTENT(OUT)       :: m
!

REAL (dp)  :: del1, del2, deltk, eps, four, one, spcmn, ten, two, zero
INTEGER    :: i, ialc, idb, ione, j, mm1, mp1, n, nmpp1, nmppq, nnn, np1
!
! IN THE INITIALIZATION WE MINIMIZE W WITH THE RESTRICTIONS
!   ABS(F*Q-WT*P) <= W AT ALL GRID POINTS WHERE F IS TO BE APPROXIMATED,
!   Q >= EPS AT ALL GRID POINTS,
!   ABS(Q(J)) <= 1.0 FOR ALL J,
!   PLUS THE OTHER CONSTRAINTS (IF ANY).
! IN THE MAIN ITERATIONS WE MINIMIZE W WITH THE RESTRICTIONS
!   (ABS(F*Q-WT*P)-DELTK*Q)/QK <= W AT ALL GRID POINTS WHERE
!   F IS TO BE APPROXIMATED,
!   ABS(Q(J)) <= 1.0 FOR ALL J,
!   PLUS THE OTHER CONSTRAINTS (IF ANY).
! EACH ABSOLUTE VALUE CONSTRAINT IS BROKEN INTO TWO CONSTRAINTS
! TO ACCOUNT FOR THE ABSOLUTE VALUE.  IN THE MAIN ITERATIONS
! THE SUBROUTINE USES THE FACT THAT THE VALUE OF QK AT GRID
! POINT I IS STORED IN V(2*I,N+1) AND DELTK IS IN
! V(2*NUMGR+1,N+1) FROM A PREVIOUS CALL TO SPQERD.
!
!***FIRST EXECUTABLE STATEMENT  SSETV2
ione=iopt-(iopt/10)*10
!
! SET MACHINE AND PRECISION DEPENDENT CONSTANTS FOR
! SUBROUTINE SSETV.
one=1.0D0
zero=one-one
two=one+one
four=two+two
ten=four+four+two

! IF WE ARE IN THE MAIN PART OF DIFFERENTIAL CORRECTION AND
! IONE IS ODD WE SET EPS=EPSQ.  IN THIS CASE WE WILL REQUIRE
! THAT THE DENOMINATOR BE >= EPS AT EVERY GRID POINT.
IF (mode > 0) THEN
  IF (ione-(ione/2)*2 <= 0) GO TO 80
  eps=epsq
ELSE

! HERE WE ARE IN THE INITIALIZATION AND WE AT LEAST TEMPORARILY
! COMPUTE EPS=MAX(0.0001,1000.0*SPCMN)
  spcmn=EPSILON(1.0_dp)
  eps=spcmn*ten**3
  IF (eps-ten**(-4) < 0.0) THEN
    eps=ten**(-4)
  END IF

! IF IONE IS ODD HERE WE INCREASE EPS TO EPSQ IF NECESSARY.
  IF (ione-(ione/2)*2 > 0) THEN
    IF (eps < epsq) eps = epsq
  END IF
END IF
! END OF SETTING MACHINE AND PRECISION DEPENDENT CONSTANTS FOR SSETV.
!
80 nmpp1=nump+1
nmppq=nump+numq
n=nmppq+1
np1=n+1
deltk=v(2*numgr+1,np1)

! M WILL KEEP TRACK OF WHICH ROW OF V IS BEING SET.  AT THE
! CONCLUSION OF SSETV M WILL BE THE TOTAL NUMBER OF CONSTRAINTS.
m=0
!
! SET THE UPPER CONSTRAINTS, IF THERE ARE ANY, EXCEPT FOR
! THE LAST COLUMN.  THEY WILL BE OF THE FORM P-Q*UP <= 0.0
IF (iopt-(iopt/1000)*1000-100 >= 0) THEN

  DO  i=1,numgr
! INDUP(I)=1 IF THERE IS AN UPPER CONSTRAINT AT GRID POINT I,
! AND INDUP(I)=0 OTHERWISE.
    IF (indup(i) <= 0) CYCLE
    m=m+1
    DO  j=1,nump
      v(m,j)=pwr(i,j)
    END DO
    DO  j=nmpp1,nmppq
      v(m,j)=-uptbl(i)*pwr(i,j)
    END DO
    v(m,n)=zero
  END DO
END IF
!
! SET THE LOWER CONSTRAINTS, IF THERE ARE ANY, EXCEPT FOR
! THE LAST COLUMN.  THEY WILL BE OF THE FORM -P+Q*LW <= 0.0
IF (iopt-(iopt/100)*100-10 >= 0) THEN

  DO  i=1,numgr
! INDLW(I)=1 IF THERE IS A LOWER CONSTRAINT AT GRID POINT I,
! AND INDLW(I)=0 OTHERWISE.
    IF (indlw(i) <= 0) CYCLE
    m=m+1
    DO  j=1,nump
      v(m,j)=-pwr(i,j)
    END DO
    DO  j=nmpp1,nmppq
      v(m,j)=altbl(i)*pwr(i,j)
    END DO
    v(m,n)=zero
  END DO
END IF
!
IF (mode > 0) GO TO 270
!
! HERE WE ARE IN THE INITIALIZATION PHASE, AND FOR EACH GRID
! POINT AT WHICH F IS TO BE APPROXIMATED WE SET THE CONSTRAINTS
! -WT*P+F*Q-W <= 0.0 AND WT*P-F*Q-W <= 0.0
DO  i=1,numgr
! INDF(I)=1 IF F IS TO BE APPROXIMATED AT GRID POINT I, AND
! INDF(I)=0 IF F IS TO BE IGNORED AT GRID POINT I.
  IF (indf(i) <= 0) CYCLE
  m=m+2
  mm1=m-1
  DO  j=1,nump
    v(mm1,j)=-wtble(i)*pwr(i,j)
    v(m,j)=-v(mm1,j)
  END DO
  DO  j=nmpp1,nmppq
    v(mm1,j)=ftbwt(i)*pwr(i,j)
    v(m,j)=-v(mm1,j)
  END DO
  v(mm1,n)=-one
  v(m,n)=-one
END DO

! SINCE WE ARE IN THE INITIALIZATION, COLUMN N+1 IS NOT NEEDED
! FOR STORAGE OF QK, SO ZERO IT OUT FOR THOSE ROWS ALREADY SET.
v(1:m,np1)=zero
GO TO 330
!
! HERE WE ARE IN THE MAIN DIFFERENTIAL CORRECTION PHASE, AND
! FOR EACH GRID POINT AT WHICH F IS TO BE APPROXIMATED SET
! THE CONSTRAINTS -WT*P+(-DELTK+F)*Q-QK*W<=0.0
! AND WT*P+(-DELTK-F)*Q-QK*W<=0.0
! (EXCEPT FOR THE LAST COLUMN).

270 DO  i=1,numgr
! INDF(I)=1 IF F IS TO BE APPROXIMATED AT GRID POINT I,
! AND INDF(I)=0 IF F IS TO BE IGNORED AT GRID POINT I.
  IF (indf(i) <= 0) CYCLE
  m=m+2
  mm1=m-1
  DO  j=1,nump
    v(mm1,j)=-wtble(i)*pwr(i,j)
    v(m,j)=-v(mm1,j)
  END DO
  del1=-deltk + ftbwt(i)
  del2=-deltk - ftbwt(i)
  DO  j=nmpp1,nmppq
    v(mm1,j)=del1*pwr(i,j)
    v(m,j)=del2*pwr(i,j)
  END DO

! QK AT GRID POINT I IS STORED IN V(2*I,N+1) FROM AN EARLIER CALL TO SPQERD.
  idb=2*i
  v(mm1,n)=-v(idb,np1)
  v(m,n)=-v(idb,np1)
END DO

! NOW THE LAST COLUMN IS NO LONGER NEEDED FOR THE STORAGE OF
! QK, SO ZERO IT OUT FOR THOSE ROWS ALREADY SET.
v(1:m,np1)=zero

! IF THE ONES DIGIT OF IOPT IS ODD HERE WE FORCE Q >= EPS.
IF (ione-(ione/2)*2 <= 0) GO TO 370
!
! HERE WE FORCE Q >= EPS.
330 DO  i=1,numgr
  m=m+1
  v(m,1:n)=zero
  DO  j=nmpp1,nmppq
    v(m,j)=-pwr(i,j)
  END DO
  v(m,np1)=-eps
END DO
!
! NOW INCLUDE THE ADDITIONAL INEQUALITY CONSTRAINTS FROM ALC,
! IF THERE ARE ANY.
370 IF (mineq > 0) THEN
  ialc=meq
  DO  i=1,mineq
    m=m+1
    ialc=ialc+1
    DO  j=1,nmppq
      v(m,j)=alc(ialc,j)
    END DO
    v(m,n)=zero
    v(m,np1)=alc(ialc,n)
  END DO
END IF
!
! NOW SET THE CONSTRAINTS OF THE FORM -Q(J) <= 1.0 AND Q(J) <= 1.0
DO  i=1,numq
  m=m+2
  mm1=m-1
  DO  j=1,n
    v(mm1,j)=zero
    v(m,j)=zero
  END DO
  nnn=nump+i
  v(mm1,nnn)=-one
  v(m,nnn)=one
  v(mm1,np1)=one
  v(m,np1)=one
END DO
!
! NOW SET THE BOTTOM ROW.  TO MINIMIZE W WE MAXIMIZE -W.
mp1=m+1
v(mp1,1:np1)=zero
v(mp1,n)=one

RETURN
END SUBROUTINE ssetv




SUBROUTINE eqelm (nvar, meq, tolel, EQ, nscol, iresl, meqa, iflag)
!***BEGIN PROLOGUE  EQELM
!***REFER TO  CDFCOR
!***ROUTINES CALLED  (NONE)
!***PURPOSE  THIS SUBROUTINE ATTEMPTS TO SOLVE FOR SOME OF THE VARIABLES IN
!            THE MEQ EQUALITY CONSTRAINTS IN TERMS OF THE OTHER VARIABLES.
!***END PROLOGUE  EQELM
!

INTEGER, INTENT(IN)        :: nvar
INTEGER, INTENT(IN)        :: meq
REAL (dp), INTENT(IN)      :: tolel
REAL (dp), INTENT(IN OUT)  :: EQ(15,16)
INTEGER, INTENT(OUT)       :: nscol(15)
INTEGER, INTENT(IN OUT)    :: iresl(30)
INTEGER, INTENT(OUT)       :: meqa
INTEGER, INTENT(OUT)       :: iflag
!

!
! GIVEN THE EQUATIONS
!   EQ(I,1)*X(1)+...+EQ(I,NVAR)*X(NVAR) = EQ(I,NVAR+1)
! FOR I=1,...,MEQ, WE CHOOSE RESOLVENTS BY TOTAL PIVOTING IN THE
! LEFT MEQ BY NVAR PORTION OF EQ, STORE THE LTH RESOLVENT
! POSITION IN (IRESL(2*L-1),IRESL(2*L)), AND APPLY ROW
! OPERATIONS TO MAKE 1.0 IN THE RESOLVENT POSITION AND 0.0
! ELSEWHERE IN THE RESOLVENT COLUMN.  ONLY RESOLVENTS WITH ABSOLUTE
! VALUE > TOLEL ARE ACCEPTED.  MEQA (WHICH STANDS FOR MEQ ACTUAL)
! WILL BE THE NUMBER OF ACCEPTABLE RESOLVENTS FOUND.  IF
! MEQA < MEQ, AND IF SOME ROW NOT CONTAINING A RESOLVENT
! HAS LAST ELEMENT > TOLEL IN ABSOLUTE VALUE, WE SET
! IFLAG=7777 AS AN INCONSISTENCY WARNING AND RETURN.  IF
! MEQA < MEQ BUT THERE IS NO INCONSISTENCY THEN THE MEQ-MEQA
! ROWS OF EQ NOT CONTAINING RERSOLVENTS WILL BE IGNORED (THIS IS
! THE REDUNDANT CONSTRAINTS SITUATION).
! NSCOL (WHICH STANDS FOR NON SOLVED FOR COLUMN) IS DEFINED BY
! NSCOL(K) = THE SUBSCRIPT OF THE KTH UNSOLVED FOR VARIABLE
! (THAT IS, VARIABLE WITH NO RESOLVENT IN ITS COLUMN) FOR K=1,...,
! NVAR-MEQA.  THUS FOR K=1,...,MEQA WE WILL HAVE
!   X(IRESL(2*K)) = -EQ(IRESL(2*K-1),NSCOL(1))*X(NSCOL(1))-...-
!   EQ(IRESL(2*K-1),NSCOL(NVAR-MEQA))*X(NSCOL(NVAR-MEQA))+
!   EQ(IRESL(2*K-1),NVAR+1).
!

REAL (dp)  :: amax, fact, one, resol, zero
INTEGER    :: i, iarg, imax, j, jmax, k, l, n, nresl

! SET PRECISION DEPENDENT CONSTANTS FOR EQELM.
!***FIRST EXECUTABLE STATEMENT  EQELM
one=1.0D0
zero=one-one
! END OF SETTING PRECISION DEPENDENT CONSTANTS FOR EQELM.
!
n=nvar+1
nresl=0
meqa=meq

! IF MEQ IS NOT POSITIVE THERE IS NOTHING TO DO HERE AND WE RETURN.
IF (meq <= 0) RETURN

! ZERO OUT NSCOL.  LATER EACH ELEMENT OF NSCOL CORRESPONDING TO A
! COLUMN CONTAINING A RESOLVENT WILL BE SET TO 1, AND NSCOL WILL
! BE READJUSTED TO ITS FINAL STATE AT THE END OF THIS SUBROUTINE
! IF THE EQUATIONS IN EQ ARE CONSISTENT.
nscol(1:nvar)=0
!
! WE ATTEMPT TO FIND MEQ RESOLVENTS USING TOTAL PIVOTING.
DO  l=1,meq
  amax=zero
  loop100:  &
  DO  i=1,meq
    loop90: &
    DO  j=1,nvar

! WHILE SEARCHING, IGNORE ROWS AND COLUMNS CONTAINING AN EARLIER RESOLVENT.
      DO  k=1,nresl
        IF (i == iresl(2*k-1)) CYCLE loop100
        IF (j == iresl(2*k)) CYCLE loop90
      END DO

! HERE LOCATION (I,J) IS NOT IN A ROW OR COLUMN WITH AN EARLIER RESOLVENT.
      IF (ABS(EQ(i,j)) > amax) THEN
        amax=ABS(EQ(i,j))
        imax=i
        jmax=j
      END IF
    END DO loop90
  END DO loop100

  IF (amax <= tolel) GO TO 170
!
! HERE WE HAVE FOUND AN ACCEPTABLE RESOLVENT.
  nresl=nresl+1
  iresl(2*nresl-1)=imax
  iresl(2*nresl)=jmax
  nscol(jmax)=1

! NOW DO ROW OPERATIONS TO MAKE 1.0 IN THE (IMAX,JMAX) POSITION
! AND 0.0 ELSEWHERE IN COLUMN JMAX (THUS SOLVING FOR X(JMAX)
! IN TERMS OF THE OTHER VARIABLES).
! FIRST DIVIDE ROW IMAX BY THE RESOLVENT.
  resol=EQ(imax,jmax)
  DO  j=1,n
    EQ(imax,j)=EQ(imax,j)/resol
  END DO
  DO  i=1,meq
    IF (i == imax) CYCLE
    fact=EQ(i,jmax)
    DO  j=1,n
      EQ(i,j)=EQ(i,j) - fact*EQ(imax,j)
    END DO
  END DO
END DO
GO TO 230
!
! HERE WE WERE UNABLE TO FIND AN ACCEPTABLE RESOLVENT.
! RESET MEQA TO THE NUMBER OF RESOLVENTS FOUND AND CHECK TO
! SEE IF ANY OF THE RIGHT SIDES OF ROWS NOT CONTAINING A
! RESOLVENT ARE > TOLEL IN ABSOLUTE VALUE.  IF SO, THE
! EQUATIONS IN EQ WERE APPARENTLY INCONSISTENT, AND IF NOT,
! THEY WERE APPARENTLY REDUNDANT.
170 meqa=nresl
loop190: &
DO  i=1,meq
  DO  k=1,nresl
    IF (i == iresl(2*k-1)) CYCLE loop190
  END DO

! HERE ROW I DOES NOT CONTAIN A RESOLVENT.
  IF (ABS(EQ(i,n)) > tolel) GO TO 220
END DO loop190
GO TO 230
!
! HERE THE CONSTRAINTS WERE INCONSISTENT.
220 iflag=7777
RETURN
!
! HERE THE CONSTRAINTS WERE CONSISTENT, AND WE SET NSCOL AND RETURN.
230 iarg=0
DO  j=1,nvar
  IF (nscol(j) > 0) CYCLE
  iarg=iarg+1
  nscol(iarg)=j
END DO

RETURN
END SUBROUTINE eqelm




SUBROUTINE scrch (m, nvar, meqa, nscol, iresl, EQ, v)
!***BEGIN PROLOGUE  SCRCH
!***REFER TO  CDFCOR
!***ROUTINES CALLED  (NONE)
!***PURPOSE  THIS SUBROUTINE USES THE NONREDUNDANT EQUALITY
!            CONSTRAINTS TO ELIMINATE MEQA VARIABLES, THUS
!            SCRUNCHING V FROM MP1 BY NP1 TO MP1 BY NP1-MEQA.
!***END PROLOGUE  SCRCH
!

INTEGER, INTENT(IN)     :: m
INTEGER, INTENT(IN)     :: nvar
INTEGER, INTENT(IN)     :: meqa
INTEGER, INTENT(IN)     :: nscol(15)
INTEGER, INTENT(IN)     :: iresl(30)
REAL (dp), INTENT(IN)   :: EQ(15,16)
REAL (dp), INTENT(OUT)  :: v(635,17)
!

REAL (dp)  :: fact
INTEGER    :: i, j, jcol, l, lcol, lrow, mp1, n, nmeq, np1, npmeq, ns
!
! IF MEQA=0 THERE IS NO SCRUNCHING TO BE DONE, AND WE RETURN.
!***FIRST EXECUTABLE STATEMENT  SCRCH
IF (meqa <= 0) RETURN
!
n=nvar+1
np1=n+1
mp1=m+1
ns=nvar-meqa
!
DO  l=1,meqa
! THE RESOLVENT IS AT POSITION (LROW,LCOL) OF EQ.
  lrow=iresl(2*l-1)
  lcol=iresl(2*l)

! IF ALL THE VARIABLES X(1),...,X(NVAR) ARE SOLVED FOR WE GO TO COLUMN NP1.
  IF (ns > 0) THEN
    DO  j=1,ns
      jcol=nscol(j)
      fact=EQ(lrow,jcol)

! WE REPLACE COLUMN JCOL OF V BY ITSELF MINUS EQ(LROW,JCOL) TIMES COLUMN LCOL
! OF V.
      DO  i=1,mp1
        v(i,jcol)=v(i,jcol) - fact*v(i,lcol)
      END DO
    END DO
  END IF
!
! NOW DO COLUMN NP1 OF V.
  fact=EQ(lrow,n)
  DO  i=1,mp1
    v(i,np1)=v(i,np1) - fact*v(i,lcol)
  END DO
END DO
!
! NOW COUMNS IRESL(2),...,IRESL(2*MEQA) OF V ARE NO LONGER
! NEEDED, AND WE COMPACTIFY V.
DO  j=1,ns
  jcol=nscol(j)
! COLUMN JCOL OF V IS NOW MOVED TO THE COLUMN J POSITION IF IT
! IS NOT ALREADY THERE.
  IF (jcol /= j) THEN
    DO  i=1,mp1
      v(i,j)=v(i,jcol)
    END DO
  END IF
END DO

! NOW MOVE THE LAST TWO COLUMNS OF V TO THE LEFT MEQA SPACES.
! NOTE THAT THE VARIABLE W CORRESPONDING TO COLUMN N = NVAR+1
! OF V IS AUTOMATICALLY AN UNSOLVED FOR VARIABLE SINCE IT IS
! NOT INCLUDED IN EQ.
nmeq=n-meqa
npmeq=nmeq+1
DO  i=1,mp1
  v(i,nmeq)=v(i,n)
  v(i,npmeq)=v(i,np1)
END DO

RETURN
END SUBROUTINE scrch




SUBROUTINE expnd (nvar, meqa, nscol, iresl, EQ, x)
!***BEGIN PROLOGUE  EXPND
!***REFER TO  CDFCOR
!***ROUTINES CALLED  (NONE)
!***PURPOSE  THIS SUBROUTINE COMPUTES THE VALUES OF THE MEQA
!            SOLVED FOR VARIABLES AND SPLICES THEM INTO X.
!***END PROLOGUE  EXPND
!

INTEGER, INTENT(IN)     :: nvar
INTEGER, INTENT(IN)     :: meqa
INTEGER, INTENT(IN)     :: nscol(15)
INTEGER, INTENT(IN)     :: iresl(30)
REAL (dp), INTENT(IN)   :: EQ(15,16)
REAL (dp), INTENT(OUT)  :: x(16)
!

INTEGER  :: j, jcol, jj, l, lcol, lrow, n, nmeq, ns
!
! IF MEQA=0 NO EXPANDING NEEDS TO BE DONE AND WE RETURN.
IF (meqa <= 0) THEN
  RETURN
END IF
!
n=nvar+1

! SHIFT THE LAST ELEMENT IN X INTO ITS PROPER PLACE.  NOTE
! THAT THIS ELEMENT CORRESPONDS TO THE VARIABLE W, WHICH IS
! UNSOLVED FOR SINCE IT IS NOT INCLUDED IN EQ.
nmeq=n-meqa
x(n)=x(nmeq)
ns=nvar-meqa

! SHIFT THE NS UNSOLVED FOR VARIABLES (NOT COUNTING W) INTO THE
! PROPER PLACES IN X, IF THERE ARE ANY.
IF (ns > 0) THEN
  DO  j=1,ns
    jj=ns-j+1
    x(nscol(jj))=x(jj)
  END DO
END IF
!
! NOW COMPUTE THE MEQA SOLVED FOR VARIABLES.
DO  l=1,meqa

! THE LTH RESOLVENT POSITION IN EQ IS (LROW,LCOL).  WE COMPUTE X(LCOL).
  lrow=iresl(2*l-1)
  lcol=iresl(2*l)
  x(lcol)=EQ(lrow,n)
  DO  j=1,ns
    jcol=nscol(j)
    x(lcol)=x(lcol) - EQ(lrow,jcol)*x(jcol)
  END DO
END DO

RETURN
END SUBROUTINE expnd




SUBROUTINE nrmlz (nump, numq, meq, mineq, nscol, iresl, alc, tolnr, x)
!***BEGIN PROLOGUE  NRMLZ
!***REFER TO  CDFCOR
!***ROUTINES CALLED  (NONE)
!***PURPOSE  THIS SUBROUTINE ATTEMPTS TO NORMALIZE X SO THAT
!            MAX(ABS(Q(J)))=1.0, OR SOME INEQUALITY CONSTRAINT
!            IN ALC WITH LEFT AND RIGHT SIDES > TOLNR HOLDS
!            WITH EQUALITY.  THE NORMALIZATION IS SKIPPED IF IT
!            WOULD REQUIRE MULTIPLYING BY A FACTOR LESS THAN 1.0,
!            OR IF SOME EQUALITY CONSTRAINT HAS RIGHT SIDE WITH
!            ABSOLUTE VALUE > TOLNR.
!***END PROLOGUE  NRMLZ
!

INTEGER, INTENT(IN)        :: nump
INTEGER, INTENT(IN)        :: numq
INTEGER, INTENT(IN)        :: meq
INTEGER, INTENT(IN)        :: mineq
INTEGER, INTENT(IN OUT)    :: nscol(15)
INTEGER, INTENT(IN OUT)    :: iresl(30)
REAL (dp), INTENT(IN)      :: alc(101,16)
REAL (dp), INTENT(IN)      :: tolnr
REAL (dp), INTENT(IN OUT)  :: x(16)
!

REAL (dp)  :: alft, amax, fact, one
INTEGER    :: i, j, mpi, n, nmppq, npj
!
! SET PRECISION DEPENDENT CONSTANTS FOR NRMLZ.
!***FIRST EXECUTABLE STATEMENT  NRMLZ
one=1.0D0
! END OF SETTING PRECISION DEPENDENT CONSTANTS FOR NRMLZ.
!
nmppq=nump+numq
n=nmppq+1

! IF THE RIGHT SIDE OF SOME EQUALITY CONSTRAINT IS > TOLNR,
! WE SKIP THE NORMALIZATION.
DO  i=1,meq
  IF (ABS(alc(i,n)) > tolnr) GO TO 150
END DO
!
! COMPUTE AMAX = MAX(TOLNR,ABS(Q(1)),...,ABS(Q(NUMQ))).
amax=tolnr
DO  j=1,numq
  npj=nump+j
  IF (ABS(x(npj)) > amax) amax=ABS(x(npj))
END DO
fact=one/amax
!
! CHECK THE INEQUALITY CONSTRAINTS IN ALC WITH SIGNIFICANTLY
! POSITIVE LEFT AND RIGHT SIDES.  MULTIPLYING ALL COEFFICIENTS
! BY A FACTOR > 1.0 WILL NOT AFFECT INEQUALITY CONSTRAINTS
! WITH LEFT AND RIGHT SIDES OF OPPOSITE SIGN OR ZERO RIGHT SIDES
! AND CAN ONLY HELP INEQUALITY CONSTRAINTS WITH NEGATIVE LEFT
! AND RIGHT SIDES.
DO  i=1,mineq
  mpi=meq+i
  IF (alc(mpi,n) <= tolnr) CYCLE
  alft=alc(mpi,1)*x(1)
  DO  j=2,nmppq
    alft=alft + alc(mpi,j)*x(j)
  END DO
  IF (alft <= tolnr) CYCLE
  IF (alc(mpi,n)/alft < fact) fact=alc(mpi,n)/alft
END DO
!
! IF FACT > 1.0, MULTIPLY ALL COEFFICIENTS BY FACT.
IF (fact > one) THEN
  DO  j=1,nmppq
    x(j)=fact*x(j)
  END DO
END IF

150 RETURN
END SUBROUTINE nrmlz




FUNCTION icnck (iopt, nump, numq, numgr, indlw, altbl, indup,  &
                uptbl, epsq, x, v, meq, mineq, alc, epres) RESULT(ival)
!***BEGIN PROLOGUE  ICNCK
!***REFER TO  CDFCOR
!***ROUTINES CALLED  (NONE)
!***PURPOSE  THIS SUBPROGRAM CHECKS ALL CONSTRAINTS, RETURNING
!            WITH VALUE 0 IF ALL ARE OK AND VALUE 1 IF ANY ARE
!            NOT SATISFIED WITHIN EPRES.
!***END PROLOGUE  ICNCK
!

INTEGER, INTENT(IN)    :: iopt
INTEGER, INTENT(IN)    :: nump
INTEGER, INTENT(IN)    :: numq
INTEGER, INTENT(IN)    :: numgr
INTEGER, INTENT(IN)    :: indlw(101)
REAL (dp), INTENT(IN)  :: altbl(101)
INTEGER, INTENT(IN)    :: indup(101)
REAL (dp), INTENT(IN)  :: uptbl(101)
REAL (dp), INTENT(IN)  :: epsq
REAL (dp), INTENT(IN)  :: x(16)
REAL (dp), INTENT(IN)  :: v(635,17)
INTEGER, INTENT(IN)    :: meq
INTEGER, INTENT(IN)    :: mineq
REAL (dp), INTENT(IN)  :: alc(101,16)
REAL (dp), INTENT(IN)  :: epres
INTEGER                :: ival
!

REAL (dp)  :: chk, one, ope
INTEGER    :: i, ione, j, l, meql, n, nmppq, np1, npj
!
! SET PRECISION DEPENDENT CONSTANTS FOR FUNCTION ICNCK.
!***FIRST EXECUTABLE STATEMENT  ICNCK
one=1.0D0
! END OF SETTING PRECISION DEPENDENT CONSTANTS FOR ICNCK.
!
nmppq=nump+numq
n=nmppq+1
np1=n+1
ione=iopt-(iopt/10)*10
!
! IF V(2*NUMGR+1,N+1) IS NEGATIVE, THEN SPQERD DETECTED A POINT
! WHERE THE DENOMINATOR WAS VERY SMALL OR NEGATIVE, SO WE SET
! ICNCK=1 AND RETURN.
IF (v(2*numgr+1,np1) < 0.0) GO TO 220
!
! CHECK LOWER RESTRICTED RANGE CONSTRAINTS, IF THERE ARE ANY.
IF (iopt-(iopt/100)*100-10 >= 0) THEN
  DO  i=1,numgr
    IF (indlw(i) <= 0) CYCLE
    IF (v(2*i-1,np1)/v(2*i,np1) - altbl(i) + epres < 0.0) GO TO 220
  END DO
END IF
!
! CHECK UPPER RESTRICTED RANGE CONSTRAINTS, IF THERE ARE ANY.
IF (iopt-(iopt/1000)*1000-100 >= 0) THEN
  DO  i=1,numgr
    IF (indup(i) > 0) THEN
      IF (v(2*i-1,np1)/v(2*i,np1) > uptbl(i)+epres) GO TO 220
    END IF
  END DO
END IF
!
! CHECK DENOMINATOR LOWER BOUND CONSTRAINTS, IF THERE ARE ANY.
IF (ione > (ione/2)*2) THEN
  DO  i=1,numgr
    IF (v(2*i,np1)-epsq+epres < 0.0) GO TO 220
  END DO
END IF
!
! CHECK ADDITIONAL EQUALITY CONSTRAINTS, IF THERE ARE ANY.
DO  l=1,meq
  chk=-alc(l,n)
  DO  j=1,nmppq
    chk=chk + alc(l,j)*x(j)
  END DO
  IF (ABS(chk) > epres) GO TO 220
END DO
!
! CHECK ADDITIONAL INEQUALITY CONSTRAINTS, IF THERE ARE ANY.
DO  l=1,mineq
  meql=meq+l
  chk=-alc(meql,n)
  DO  j=1,nmppq
    chk=chk + alc(meql,j)*x(j)
  END DO
  IF (chk > epres) GO TO 220
END DO
!
! CHECK THE DENOMINATOR COEFFICIENT BOUNDS.
ope=one+epres
DO  j=1,numq
  npj=nump+j
  IF (ABS(x(npj)) > ope) GO TO 220
END DO
!
! IF WE REACH THIS SPOT, THEN ALL CONSTRAINTS ARE OK WITHIN EPRES.
ival=0
RETURN
!
! HERE AT LEAST ONE CONSTRAINT WAS NOT SATISFIED WITHIN EPRES.
220 ival=1

RETURN
END FUNCTION icnck

END MODULE Best_rational
